"""Wrapper for performing an electronic circular dichroism (ECD)
calculation."""

import numpy as np

from pymolresponse.constants import HARTREE_TO_EV, HARTREE_TO_INVCM, alpha, esuecd
from pymolresponse.core import Program
from pymolresponse.molecular_property import TransitionProperty
from pymolresponse.operators import Operator
from pymolresponse.td import TDHF
from pymolresponse.utils import form_indices_zero


class ECD(TransitionProperty):
    """Wrapper for performing an electronic circular dichroism (ECD)
    calculation.
    """

    def __init__(
        self,
        program: Program,
        program_obj,
        driver: TDHF,
        mocoeffs: np.ndarray,
        moenergies: np.ndarray,
        occupations: np.ndarray,
        *,
        do_tda: bool = False,
        do_dipvel: bool = False,
    ) -> None:
        super().__init__(
            program, program_obj, driver, mocoeffs, moenergies, occupations, do_tda=do_tda
        )
        self.do_dipvel = do_dipvel

    def form_operators(self) -> None:
        if self.program == Program.PySCF:
            from pymolresponse.interfaces.pyscf import integrals

            integral_generator = integrals.IntegralsPyscf(self.program_obj)
        elif self.program == Program.Psi4:
            from pymolresponse.interfaces.psi4 import integrals

            integral_generator = integrals.IntegralsPsi4(self.program_obj)
        else:
            raise RuntimeError

        operator_angmom = Operator(
            label="angmom", is_imaginary=True, is_spin_dependent=False, triplet=False
        )
        operator_angmom.ao_integrals = integral_generator.integrals(integrals.ANGMOM_COMMON_GAUGE)
        self.driver.add_operator(operator_angmom)

        operator_diplen = Operator(
            label="dipole", is_imaginary=False, is_spin_dependent=False, triplet=False
        )
        operator_diplen.ao_integrals = integral_generator.integrals(integrals.DIPOLE)
        self.driver.add_operator(operator_diplen)

        if self.do_dipvel:
            operator_dipvel = Operator(
                label="dipvel", is_imaginary=True, is_spin_dependent=False, triplet=False
            )
            operator_dipvel.ao_integrals = integral_generator.integrals(integrals.DIPVEL)
            self.driver.add_operator(operator_dipvel)

    def form_results(self) -> None:
        operator_angmom = self.driver.solver.operators[0]
        operator_diplen = self.driver.solver.operators[1]
        assert len(operator_angmom.transition_moments) == len(operator_diplen.transition_moments)
        nstates = len(operator_diplen.transition_moments)
        rotational_strengths_diplen = []
        rotational_strengths_dipvel = []
        if self.do_dipvel:
            assert len(self.driver.solver.operators) == 3
            operator_dipvel = self.driver.solver.operators[2]
            assert len(operator_dipvel.transition_moments) == nstates
        for stateidx in range(nstates):
            print("-" * 78)
            eigval = self.driver.solver.eigvals[stateidx].real
            rotstr_diplen = (
                esuecd
                * (-1 / 2)
                * np.dot(
                    operator_diplen.transition_moments[stateidx],
                    operator_angmom.transition_moments[stateidx],
                )
            )
            print("length  ", rotstr_diplen)
            rotational_strengths_diplen.append(rotstr_diplen)
            if self.do_dipvel:
                rotstr_dipvel = (
                    esuecd
                    * (-1 / 2)
                    * np.dot(
                        operator_dipvel.transition_moments[stateidx] / eigval,
                        operator_angmom.transition_moments[stateidx],
                    )
                )
                print("velocity", rotstr_dipvel)
                rotational_strengths_dipvel.append(rotstr_dipvel)
        self.rotational_strengths_diplen = np.array(rotational_strengths_diplen)
        if self.do_dipvel:
            self.rotational_strengths_dipvel = np.array(rotational_strengths_dipvel)

    def print_results_nwchem(self) -> str:
        excitation_block = self.driver.print_results_nwchem()
        lines = [excitation_block]
        energies = self.driver.solver.eigvals.real
        energies_ev = energies * HARTREE_TO_EV
        op_diplen = self.driver.solver.operators[1]
        tmom_diplen = op_diplen.transition_moments
        etoscslen = op_diplen.total_oscillator_strengths
        op_angmom = self.driver.solver.operators[0]
        tmom_angmom = op_angmom.transition_moments
        rotstrlen = self.rotational_strengths_diplen
        if self.do_dipvel:
            op_dipvel = self.driver.solver.operators[2]
            rotstrvel = self.rotational_strengths_dipvel
            tmom_dipvel = op_dipvel.transition_moments
            etoscsvel = op_dipvel.total_oscillator_strengths
        nstates = len(energies)
        for state in range(nstates):
            lines.append(
                "  ----------------------------------------------------------------------------"
            )
            lines.append(
                f"  Root {state + 1:>3d} singlet a{energies[state]:>25.9f} a.u.{energies_ev[state]:>22.4f} eV"
            )
            lines.append(
                "  ----------------------------------------------------------------------------"
            )
            lines.append(
                f"     Transition Moments    X{tmom_diplen[state, 0]:>9.5f}   Y{tmom_diplen[state, 1]:>9.5f}   Z{tmom_diplen[state, 2]:>9.5f}"
            )
            ## TODO these require second moment (length) integrals
            ## lines.append(f'     Transition Moments   XX -0.28379  XY  0.08824  XZ -0.17416')
            ## lines.append(f'     Transition Moments   YY -0.40247  YZ -0.45981  ZZ  0.59211')
            lines.append(f"     Dipole Oscillator Strength {etoscslen[state]:>31.5f}")
            lines.append("")
            lines.append("     Electric Transition Dipole:")
            lines.append(
                f"            X{tmom_diplen[state, 0]:>13.7f}   Y{tmom_diplen[state, 1]:>13.7f}   Z{tmom_diplen[state, 2]:>13.7f}"
            )
            lines.append("     Magnetic Transition Dipole (Length):")
            lines.append(
                f"            X{tmom_angmom[state, 0]:>13.7f}   Y{tmom_angmom[state, 1]:>13.7f}   Z{tmom_angmom[state, 2]:>13.7f}"
            )
            lines.append("     Magnetic Transition Dipole * 1/c :")
            lines.append(
                f"            X{tmom_angmom[state, 0] * alpha:>13.7f}   Y{tmom_angmom[state, 1] * alpha:>13.7f}   Z{tmom_angmom[state, 2] * alpha:>13.7f}"
            )
            lines.append(f"     Rotatory Strength (1E-40 esu**2cm**2):{rotstrlen[state]:>21.7f}")
            lines.append("")
            if self.do_dipvel:
                lines.append("     Electric Transition Dipole (velocity representation):")
                lines.append(
                    f"            X{tmom_dipvel[state, 0]:>13.7f}   Y{tmom_dipvel[state, 1]:>13.7f}   Z{tmom_dipvel[state, 2]:>13.7f}"
                )
                # lines.append(f'     Oscillator Strength (velocity repr.) :            0.0069989')
                lines.append(
                    f"     Oscillator Strength (velocity repr.) :{etoscsvel[state]:>21.7f}"
                )
                # lines.append(f'     Oscillator Strength (mixed repr.   ) :            0.0074981')
                # lines.append(f'     Oscillator Strength (mixed repr.   ) :{>21.7f}')
                lines.append(
                    f"     Rotatory Strength   (velocity repr.) :{rotstrvel[state]:>21.7f}"
                )
                lines.append("")
            # lines.append(str(self.driver.solver.eigvecs[:, state]))
            # lines.append(str(self.driver.solver.eigvecs_normed[:, state]))

        return "\n".join(lines)

    def print_results_orca(self) -> str:
        excitation_block = self.driver.print_results_orca()
        lines = [excitation_block]
        energies = self.driver.solver.eigvals.real
        energies_to_invcm = energies * HARTREE_TO_INVCM
        energies_to_nm = 10000000 / energies_to_invcm
        op_diplen = self.driver.solver.operators[1]
        etoscslen = op_diplen.total_oscillator_strengths
        tmom_diplen = op_diplen.transition_moments
        t2_diplen = np.asarray(
            [np.dot(tmom_diplen[x], tmom_diplen[x]) for x in range(len(tmom_diplen))]
        )
        rotstrlen = self.rotational_strengths_diplen
        op_angmom = self.driver.solver.operators[0]
        tmom_angmom = op_angmom.transition_moments
        if self.do_dipvel:
            op_dipvel = self.driver.solver.operators[2]
            rotstrvel = self.rotational_strengths_dipvel  # noqa: F841
            etoscsvel = op_dipvel.total_oscillator_strengths
            tmom_dipvel = op_dipvel.transition_moments
            t2_dipvel = np.asarray(
                [np.dot(tmom_dipvel[x], tmom_dipvel[x]) for x in range(len(tmom_dipvel))]
            )
        nstates = len(energies)
        lines.append(
            "-----------------------------------------------------------------------------"
        )
        lines.append("         ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS")
        lines.append(
            "-----------------------------------------------------------------------------"
        )
        lines.append("State   Energy  Wavelength   fosc         T2         TX        TY        TZ")
        lines.append("        (cm-1)    (nm)                  (au**2)     (au)      (au)      (au)")
        lines.append(
            "-----------------------------------------------------------------------------"
        )
        for state in range(nstates):
            lines.append(
                f"{state + 1:>4d}{energies_to_invcm[state]:>10.1f}{energies_to_nm[state]:>9.1f}{etoscslen[state]:>14.9f}{t2_diplen[state]:>10.5f}{tmom_diplen[state, 0]:>10.5f}{tmom_diplen[state, 1]:>10.5f}{tmom_diplen[state, 2]:>10.5f}"
            )
        lines.append("")
        lines.append(
            "-----------------------------------------------------------------------------"
        )
        lines.append("         ABSORPTION SPECTRUM VIA TRANSITION VELOCITY DIPOLE MOMENTS")
        lines.append(
            "-----------------------------------------------------------------------------"
        )
        lines.append("State   Energy  Wavelength   fosc         P2         PX        PY        PZ")
        lines.append("        (cm-1)    (nm)                  (au**2)     (au)      (au)      (au)")
        lines.append(
            "-----------------------------------------------------------------------------"
        )
        for state in range(nstates):
            lines.append(
                f"{state + 1:>4d}{energies_to_invcm[state]:>10.1f}{energies_to_nm[state]:>9.1f}{etoscsvel[state]:>14.9f}{t2_dipvel[state]:>10.5f}{tmom_dipvel[state, 0]:>10.5f}{tmom_dipvel[state, 1]:>10.5f}{tmom_dipvel[state, 2]:>10.5f}"
            )
        lines.append("")
        lines.append("-------------------------------------------------------------------")
        lines.append("                             CD SPECTRUM")
        lines.append("-------------------------------------------------------------------")
        lines.append("State   Energy Wavelength       R         MX        MY        MZ")
        lines.append("        (cm-1)   (nm)       (1e40*cgs)   (au)      (au)      (au)")
        lines.append("-------------------------------------------------------------------")
        for state in range(nstates):
            lines.append(
                f"{state + 1:>4d}{energies_to_invcm[state]:>10.1f}{energies_to_nm[state]:>9.1f}{rotstrlen[state]:>13.5f}{tmom_angmom[state, 0]:>10.5f}{tmom_angmom[state, 1]:>10.5f}{tmom_angmom[state, 2]:>10.5f}"
            )
        lines.append("")
        return "\n".join(lines)

    _HAMILTONIAN_PREFIX_QCHEM = {"tda": "", "rpa": "X: "}

    # TODO cutoff taken from ORCA, check the source code to see the
    # real criterion
    def print_results_qchem(self, cutoff: float = 0.01) -> str:
        energies = self.driver.solver.eigvals.real
        energies_ev = energies * HARTREE_TO_EV
        op_diplen = self.driver.solver.operators[1]
        tmom_diplen = op_diplen.transition_moments
        etoscslen = op_diplen.total_oscillator_strengths
        nocc_tot, nvirt_tot, _, _ = self.driver.solver.occupations
        indices = form_indices_zero(nocc_tot, nvirt_tot)
        eigvecs = self.driver.solver.eigvecs
        square_eigvecs = np.power(eigvecs, 2)
        lines = []
        lines.append(" ---------------------------------------------------")
        lines.append(
            f"               {self.driver._HAMILTONIAN_MAP_ORCA[self.driver.hamiltonian]} Excitation Energies              "
        )
        lines.append(" ---------------------------------------------------")
        lines.append("")
        nstates = len(energies)
        for state in range(nstates):
            lines.append(
                f" Excited state{state + 1:>4d}: excitation energy (eV) ={energies_ev[state]:>10.4f}"
            )
            lines.append(f" Total energy for state{state + 1:>3d}:{0:>31.8f} au")
            lines.append(f"    Multiplicity: {self.driver._SPIN_MAP_QCHEM[self.driver.spin]}")
            lines.append(
                f"    Trans. Mom.:{tmom_diplen[state, 0]:>8.4f} X{tmom_diplen[state, 1]:>9.4f} Y{tmom_diplen[state, 2]:>9.4f} Z"
            )
            lines.append(f"    Strength   :{etoscslen[state]:>17.10f}")
            eigvec_state = eigvecs[:, state]
            square_eigvec_state = square_eigvecs[:, state]
            mask = square_eigvec_state > cutoff
            coeffs_cutoff = eigvec_state[mask]
            mask_indices = np.array([p for (p, b) in enumerate(mask) if b])
            for i in range(len(coeffs_cutoff)):
                iocc, ivirt = indices[mask_indices[i]]
                lines.append(
                    f"    {self._HAMILTONIAN_PREFIX_QCHEM[self.driver.hamiltonian]}D({iocc + 1:>3d}) --> V({ivirt + 1:>3d}) amplitude ={coeffs_cutoff[i]:>8.4f}"
                )
            lines.append("")
        return "\n".join(lines)



from typing import Tuple

import numpy as np

import pyscf

from pymolresponse.utils import fix_mocoeffs_shape


def occupations_from_pyscf_mol(
    mol: pyscf.gto.Mole, C: np.ndarray
) -> np.ndarray[Tuple[int], np.dtype[int]]:
    norb = fix_mocoeffs_shape(C).shape[-1]
    nocc_a, nocc_b = mol.nelec
    nvirt_a, nvirt_b = norb - nocc_a, norb - nocc_b
    ret = np.asarray([nocc_a, nvirt_a, nocc_b, nvirt_b], dtype=int)
    return ret



from typing import Optional, Tuple, Union

import numpy as np

import psi4

from pymolresponse.integrals import JK, IntegralLabel, Integrals

DIPOLE = object()
DIPVEL = object()
ANGMOM_COMMON_GAUGE = object()


class IntegralsPsi4(Integrals):
    def __init__(self, wfn_or_mol: Union[psi4.core.Molecule, psi4.core.Wavefunction]) -> None:
        super().__init__()

        if isinstance(wfn_or_mol, psi4.core.Molecule):
            wfn = psi4.core.Wavefunction.build(wfn_or_mol, psi4.core.get_global_option("BASIS"))
        elif isinstance(wfn_or_mol, psi4.core.Wavefunction):
            wfn = wfn_or_mol
        else:
            raise RuntimeError
        self._mints = psi4.core.MintsHelper(wfn)

    def _compute(self, label: IntegralLabel) -> np.ndarray:
        if label == DIPOLE:
            return np.stack([np.asarray(Mc) for Mc in self._mints.ao_dipole()])
        elif label == DIPVEL:
            return np.stack([np.asarray(Mc) for Mc in self._mints.ao_nabla()])
        elif label == ANGMOM_COMMON_GAUGE:
            return np.stack([np.asarray(Lc) for Lc in self._mints.ao_angular_momentum()])
        else:
            raise RuntimeError


# Taken from Psi4NumPy's helper_HF.py
def compute_jk(jk: psi4.core.JK, C_left, C_right: Optional = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    A python wrapper for a Psi4 JK object to consume and produce NumPy arrays.

    Computes the following matrices:
    D = C_left C_right.T
    J_pq = (pq|rs) D_rs
    K_pq = (pr|qs) D_rs

    Parameters
    ----------
    jk : psi4.core.JK
        A initialized Psi4 JK object
    C_left : list of array_like or a array_like object
        Orbitals used to compute the JK object with
    C_right : list of array_like (optional, None)
        Optional C_right orbitals, otherwise it is assumed C_right == C_left

    Returns
    -------
    JK : tuple of ndarray
        Returns the J and K objects

    Notes
    -----
    This function uses the Psi4 JK object and will compute the initialized JK type (DF, PK, OUT_OF_CORE, etc)


    Examples
    --------

    ndocc = 5
    nbf = 15

    Cocc = np.random.rand(nbf, ndocc)

    jk = psi4.core.JK.build(wfn.basisset())
    jk.set_memory(int(1.25e8))  # 1GB
    jk.initialize()
    jk.print_header()


    J, K = compute_jk(jk, Cocc)

    J_list, K_list = compute_jk(jk, [Cocc, Cocc])
    """

    # Clear out the matrices
    jk.C_clear()

    list_input = True
    if not isinstance(C_left, (list, tuple)):
        C_left = [C_left]
        list_input = False

    for c in C_left:
        mat = psi4.core.Matrix.from_array(c)
        jk.C_left_add(mat)

    # Do we have C_right?
    if C_right is not None:
        if not isinstance(C_right, (list, tuple)):
            C_right = [C_right]

        if len(C_left) != len(C_right):
            raise ValueError("JK: length of left and right matrices is not equal")

        if not isinstance(C_right, (list, tuple)):
            C_right = [C_right]

        for c in C_right:
            mat = psi4.core.Matrix.from_array(c)
            jk.C_right_add(mat)

    # Compute the JK
    jk.compute()

    # Unpack
    J = []
    K = []
    for n in range(len(C_left)):
        J.append(np.array(jk.J()[n]))
        K.append(np.array(jk.K()[n]))

    jk.C_clear()

    # Duck type the return
    if list_input:
        return (J, K)
    else:
        return (J[0], K[0])


class JKPsi4(JK):
    def __init__(self, wfn) -> None:
        super().__init__()

        self._jk = psi4.core.JK.build(wfn.basisset())
        self._jk.initialize()

    def compute_from_density(self, D: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError

    def compute_from_mocoeffs(
        self, C_left: np.ndarray, C_right: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        # TODO is would be good to understand why this doesn't work.
        #
        # self._jk.finalize()
        # self._jk.initialize()
        # self._jk.C_clear()
        # self._jk.C_left_add(psi4.core.Matrix.from_array(C_left))
        # if C_right is not None:
        #     if len(C_left) != len(C_right):
        #         raise ValueError(
        #             "JK: length of left and right MO coefficient matrices is not equal"
        #         )
        #     self._jk.C_right_add(psi4.core.Matrix.from_array(C_right))
        # self._jk.compute()
        # self._jk.C_clear()
        # return self._jk.J()[0].np, self._jk.K()[0].np
        return compute_jk(self._jk, C_left, C_right)



"""Drivers for solving the time-dependent Hartree-Fock (TDHF)
equations."""

import numpy as np

from pymolresponse.constants import HARTREE_TO_EV, HARTREE_TO_INVCM
from pymolresponse.core import Hamiltonian, Program, Spin
from pymolresponse.cphf import CPHF
from pymolresponse.solvers import EigSolver, EigSolverTDA, Solver
from pymolresponse.utils import form_indices_orbwin, form_vec_energy_differences


class TDHF(CPHF):
    """Driver for solving the time-dependent Hartree-Fock (TDHF)
    equations, also called the random phase approximation (RPA)
    equations.
    """

    def __init__(self, solver: Solver) -> None:
        assert isinstance(solver, EigSolver)
        super().__init__(solver)

    def run(self, hamiltonian: Hamiltonian, spin: Spin, program: Program, program_obj) -> None:
        assert isinstance(hamiltonian, Hamiltonian)
        assert isinstance(spin, Spin)
        assert isinstance(program, (Program, type(None)))
        # TODO program_obj

        self.solver.run(hamiltonian, spin, program, program_obj)
        # TODO Is there an equivalent to the uncoupled result? Just
        # orbital energy differences?
        self.form_results()

    def form_results(self) -> None:
        nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = self.solver.occupations
        nov_alph = nocc_alph * nvirt_alph
        # nov_beta = nocc_beta * nvirt_beta
        self.solver.eigvecs_normed = self.solver.eigvecs.copy()
        # This is because we've calculated all possible roots.
        for idx in range(nov_alph):
            eigvec = self.solver.eigvecs[:, idx]
            x_normed, y_normed = self.solver.norm_xy(eigvec, nocc_alph, nvirt_alph)
            eigvec_normed = np.concatenate((x_normed.flatten(), y_normed.flatten()), axis=0)
            self.solver.eigvecs_normed[:, idx] = eigvec_normed
            eigval = self.solver.eigvals[idx].real
            # contract the components of every operator with every
            # eigenvector
            for operator in self.solver.operators:
                integrals = operator.mo_integrals_ai_supervector_alph[:, :, 0]
                # TODO why the 2?
                res_normed = 2 * np.dot(integrals, eigvec_normed)
                transition_moment = res_normed
                oscillator_strength = (2 / 3) * eigval * transition_moment**2
                total_oscillator_strength = (
                    (2 / 3) * eigval * np.dot(transition_moment, transition_moment)
                )
                if not hasattr(operator, "transition_moments"):
                    operator.transition_moments = []
                if not hasattr(operator, "oscillator_strengths"):
                    operator.oscillator_strengths = []
                if not hasattr(operator, "total_oscillator_strengths"):
                    operator.total_oscillator_strengths = []
                operator.transition_moments.append(transition_moment)
                operator.oscillator_strengths.append(oscillator_strength)
                operator.total_oscillator_strengths.append(total_oscillator_strength)
        for operator in self.solver.operators:
            operator.transition_moments = np.array(operator.transition_moments)
            operator.oscillator_strengths = np.array(operator.oscillator_strengths)
            operator.total_oscillator_strengths = np.array(operator.total_oscillator_strengths)

    def print_results(self) -> None:
        energies = self.solver.eigvals.real
        for idx in range(len(energies)):
            print("=" * 78)
            print(f" State: {idx + 1}")
            print(f" Excitation energy [a.u.]: {energies[idx]}")
            print(f" Excitation energy [eV]  : {energies[idx] * HARTREE_TO_EV}")
            for operator in self.solver.operators:
                transition_moment = operator.transition_moments[idx]
                oscillator_strength = operator.oscillator_strengths[idx]
                total_oscillator_strength = operator.total_oscillator_strengths[idx]
                print("-" * 78)
                print(f" Operator: {operator.label}")
                print(f" Transition moment: {transition_moment}")
                print(f" Oscillator strength: {oscillator_strength}")
                print(f" Oscillator strength (total): {total_oscillator_strength}")

    def print_results_nwchem(self) -> str:
        # TODO UHF
        nocc_tot, nvirt_tot, _, _ = self.solver.occupations
        moene = np.diag(self.solver.moenergies[0])
        moene_occ = moene[:nocc_tot]
        moene_virt = moene[nocc_tot:]
        ediff = form_vec_energy_differences(moene_occ, moene_virt) * HARTREE_TO_EV
        idxsort = np.argsort(ediff)
        ediff_sorted = ediff[idxsort]
        indices_unrestricted_orbwin = form_indices_orbwin(nocc_tot, nvirt_tot)
        indices_sorted = [indices_unrestricted_orbwin[i] for i in idxsort]
        ndiff = 10
        lines = []
        lines.append(f"   {ndiff:>2d} smallest eigenvalue differences (eV) ")
        lines.append("--------------------------------------------------------")
        lines.append("  No. Spin  Occ  Vir  Irrep   E(Occ)    E(Vir)   E(Diff)")
        lines.append("--------------------------------------------------------")
        for idx in range(ndiff):
            iocc, ivirt = indices_sorted[idx]
            lines.append(
                f"{idx + 1:>5d}{1:>5d}{iocc + 1:>5d}{ivirt + 1:>5d} "
                f"{'X':<5}{moene[iocc]:>10.3f}{moene[ivirt]:>10.3f}{ediff_sorted[idx]:>10.3f}"
            )
        lines.append("--------------------------------------------------------")
        return "\n".join(lines)

    _HAMILTONIAN_MAP_ORCA = {Hamiltonian.TDA: "CIS-", Hamiltonian.RPA: "RPA "}

    _SPIN_MAP_ORCA = {Spin.singlet: "SINGLETS", Spin.triplet: "TRIPLETS"}

    def print_results_orca(self, cutoff: float = 0.01) -> str:
        energies = self.solver.eigvals.real
        energies_ev = energies * HARTREE_TO_EV
        energies_invcm = energies * HARTREE_TO_INVCM
        nocc_tot, nvirt_tot, _, _ = self.solver.occupations
        indices = form_indices_orbwin(nocc_tot, nvirt_tot)
        eigvecs = self.solver.eigvecs
        # eigvecs_normed = self.solver.eigvecs_normed
        square_eigvecs = np.power(eigvecs, 2)
        # square_eigvecs_normed = np.power(eigvecs_normed, 2)
        # mask = square_eigvecs > cutoff
        # print(square_eigvecs.shape)
        # print(mask.shape)
        # mask_normed = square_eigvecs_normed > cutoff
        # print(eigvecs[mask].shape)
        # print(eigvecs_normed[mask_normed])
        # print(square_eigvecs[mask].shape)
        # print(square_eigvecs_normed[mask_normed])
        lines = [
            "-----------------------------",
            f"{self._HAMILTONIAN_MAP_ORCA[self.hamiltonian]}EXCITED STATES ({self._SPIN_MAP_ORCA[self.spin]})",
            "-----------------------------",
            "",
            f"the weight of the individual excitations are printed if larger than {cutoff:>4.2f}",
            "",
        ]
        nstates = len(energies)
        for state in range(nstates):
            lines.append(
                f"STATE{state + 1:>3d}:  E={energies[state]:>11.6f} "
                f"au{energies_ev[state]:>11.3f} eV{energies_invcm[state]:>11.1f} cm**-1"
            )
            eigvec_state = eigvecs[:, state]
            square_eigvec_state = square_eigvecs[:, state]
            mask = square_eigvec_state > cutoff
            coeffs_cutoff = eigvec_state[mask]
            weight_cutoff = square_eigvec_state[mask]
            mask_indices = np.array([p for (p, b) in enumerate(mask) if b])
            for i in range(len(coeffs_cutoff)):
                iocc, ivirt = indices[mask_indices[i]]
                lines.append(
                    f"{iocc:>6d}a ->{ivirt:>4d}a  :{weight_cutoff[i]:>13.6f} (c={coeffs_cutoff[i]:>12.8f})"
                )
            lines.append("")
        return "\n".join(lines)

    _SPIN_MAP_QCHEM = {Spin.singlet: "Singlet", Spin.triplet: "Triplet"}


class TDA(TDHF):
    """Driver for solving the time-dependent Hartree-Fock equations with the
    Tamm-Dancoff approximation (TDA), also called the configuration
    interaction with single excitation (CIS) equations when no XC contribution
    is present.
    """

    def __init__(self, solver: EigSolverTDA) -> None:
        assert isinstance(solver, EigSolverTDA)
        super().__init__(solver)

    def form_results(self) -> None:
        nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = self.solver.occupations
        nov_alph = nocc_alph * nvirt_alph
        # nov_beta = nocc_beta * nvirt_beta
        self.solver.eigvecs_normed = self.solver.eigvecs.copy()
        # This is because we've calculated all possible roots.

        for idx in range(nov_alph):
            norm = 1 / np.sqrt(2)
            eigvec = self.solver.eigvecs[:, idx]
            eigvec_normed = eigvec * norm
            self.solver.eigvecs_normed[:, idx] = eigvec_normed
            eigval = self.solver.eigvals[idx].real
            # contract the components of every operator with every
            # eigenvector
            for operator in self.solver.operators:
                integrals = operator.mo_integrals_ai_alph[:, :, 0]
                # TODO why the 2?
                res_normed = 2 * np.dot(integrals, eigvec_normed)
                transition_moment = res_normed
                oscillator_strength = (2 / 3) * eigval * transition_moment**2
                total_oscillator_strength = (
                    (2 / 3) * eigval * np.dot(transition_moment, transition_moment)
                )
                if not hasattr(operator, "transition_moments"):
                    operator.transition_moments = []
                if not hasattr(operator, "oscillator_strengths"):
                    operator.oscillator_strengths = []
                if not hasattr(operator, "total_oscillator_strengths"):
                    operator.total_oscillator_strengths = []
                operator.transition_moments.append(transition_moment)
                operator.oscillator_strengths.append(oscillator_strength)
                operator.total_oscillator_strengths.append(total_oscillator_strength)
        for operator in self.solver.operators:
            operator.transition_moments = np.array(operator.transition_moments)
            operator.oscillator_strengths = np.array(operator.oscillator_strengths)
            operator.total_oscillator_strengths = np.array(operator.total_oscillator_strengths)



from typing import Sequence

import numpy as np

from pymolresponse.core import Program
from pymolresponse.cphf import CPHF
from pymolresponse.molecular_property import ResponseProperty
from pymolresponse.operators import Operator


class ORD(ResponseProperty):
    def __init__(
        self,
        program: Program,
        program_obj,
        driver: CPHF,
        mocoeffs: np.ndarray,
        moenergies: np.ndarray,
        occupations: np.ndarray,
        *,
        frequencies: Sequence[float] = [0.0],
        do_dipvel: bool = False,
    ) -> None:
        super().__init__(
            program, program_obj, driver, mocoeffs, moenergies, occupations, frequencies=frequencies
        )
        self.do_dipvel = do_dipvel

    def form_operators(self) -> None:
        if self.program == Program.PySCF:
            from pymolresponse.interfaces.pyscf import integrals

            integral_generator = integrals.IntegralsPyscf(self.program_obj)
        elif self.program == Program.Psi4:
            from pymolresponse.interfaces.psi4 import integrals

            integral_generator = integrals.IntegralsPsi4(self.program_obj)
        else:
            raise RuntimeError

        operator_angmom = Operator(
            label="angmom", is_imaginary=True, is_spin_dependent=False, triplet=False
        )
        operator_angmom.ao_integrals = integral_generator.integrals(integrals.ANGMOM_COMMON_GAUGE)  # ty: ignore[invalid-argument-type]
        self.driver.add_operator(operator_angmom)

        operator_diplen = Operator(
            label="dipole", is_imaginary=False, is_spin_dependent=False, triplet=False
        )
        operator_diplen.ao_integrals = integral_generator.integrals(integrals.DIPOLE)  # ty: ignore[invalid-argument-type]
        self.driver.add_operator(operator_diplen)

        if self.do_dipvel:
            operator_dipvel = Operator(
                label="dipvel", is_imaginary=True, is_spin_dependent=False, triplet=False
            )
            operator_dipvel.ao_integrals = integral_generator.integrals(integrals.DIPVEL)  # ty: ignore[invalid-argument-type]
            self.driver.add_operator(operator_dipvel)

    def form_results(self) -> None:
        assert len(self.driver.results) == len(self.frequencies)
        self.polarizabilities = []
        self.polarizabilities_lenmag = []

        for idxf, frequency in enumerate(self.frequencies):
            # print('=' * 78)
            results = self.driver.results[idxf]
            if self.do_dipvel:
                assert results.shape == (9, 9)
            else:
                assert results.shape == (6, 6)

            # Ordering of the operators:
            # 1. angular momentum
            # 2. dipole (length gauge)
            # 3. dipole (velocity gauge) [if calculated]
            # print('frequency')
            # print(frequency)

            polarizability = results[3:6, 3:6]
            self.polarizabilities.append(polarizability)

            # print('symmetry check')
            # abs_diff = results[0:3, 3:6] - results[3:6, 0:3].T
            # print(abs_diff)
            # print('electric dipole-magnetic dipole polarizabilities')
            # eigvals, iso, _ = tensor_printer(results[0:3, 3:6])
            # eigvals, iso, _ = tensor_printer(results[3:6, 0:3])
            polarizability_lenmag = results[0:3, 3:6]
            self.polarizabilities_lenmag.append(polarizability_lenmag)



import numpy as np

import pyscf

from pymolresponse import cphf, solvers, utils
from pymolresponse.core import Hamiltonian, Program, Spin
from pymolresponse.interfaces.pyscf import molecules
from pymolresponse.interfaces.pyscf.utils import occupations_from_pyscf_mol
from pymolresponse.properties import magnetic

# These were generated using DALTON.
ref_magnetizability_rhf = np.array(
    [
        [9.491770490066, -0.000297478459, -2.237615614426],
        [-0.000297478459, 51.486607258336, 0.010790985557],
        [-2.237615614426, 0.010790985557, 11.943759880337],
    ]
)

ref_magnetizability_rohf = np.array(
    [
        [10.730460126094, -0.031625404934, -2.412066190731],
        [-0.031625404934, 50.874378111410, -0.064676119234],
        [-2.412066190731, -0.064676119234, 13.770943967357],
    ]
)


def test_magnetizability_rhf() -> None:
    mol = molecules.molecule_glycine_sto3g()
    mol.build()

    mf = pyscf.scf.RHF(mol)
    mf.scf()

    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    calculator_common = magnetic.Magnetizability(
        Program.PySCF, mol, cphf.CPHF(solvers.ExactInv(C, E, occupations)), C, E, occupations
    )
    calculator_common.form_operators()
    calculator_common.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    calculator_common.form_results()
    ref_eigvals, ref_iso, _ = utils.tensor_printer(ref_magnetizability_rhf)
    res_eigvals, res_iso, _ = utils.tensor_printer(calculator_common.magnetizability)
    thresh_eigval = 1.0e-3
    for i in range(3):
        assert abs(ref_eigvals[i] - res_eigvals[i]) < thresh_eigval

    # TODO it isn't so simple; there are actually many more terms
    # present when using London orbitals.
    # calculator_giao = magnetic.Magnetizability(mol, C, E, occupations, hamiltonian='rpa', spin='singlet', use_giao=True)
    # calculator_giao.form_operators()
    # calculator_giao.run()
    # calculator_giao.form_results()


def test_magnetizability_uhf() -> None:
    mol = molecules.molecule_glycine_sto3g()
    mol.charge = 1
    mol.spin = 1
    mol.build()

    mf = pyscf.scf.uhf.UHF(mol)
    mf.scf()

    assert isinstance(mf.mo_coeff, np.ndarray)
    assert isinstance(mf.mo_energy, np.ndarray)
    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    calculator_common = magnetic.Magnetizability(
        Program.PySCF, mol, cphf.CPHF(solvers.ExactInv(C, E, occupations)), C, E, occupations
    )
    calculator_common.form_operators()
    calculator_common.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    calculator_common.form_results()
    ref_eigvals, ref_iso, _ = utils.tensor_printer(ref_magnetizability_rohf)
    res_eigvals, res_iso, _ = utils.tensor_printer(calculator_common.magnetizability)
    thresh_eigval = 1.0e-1
    for i in range(3):
        assert abs(ref_eigvals[i] - res_eigvals[i]) < thresh_eigval


# These were generated with this program.
ref_electronicgtensor_tiny = np.array(
    [[1.15894158e-05, 0.0, 0.0], [0.0, 1.15894158e-05, 0.0], [0.0, 0.0, 0.0]]
)
ref_electronicgtensor_small = np.array(
    [
        [2.68449088e-03, -7.88488679e-04, -1.71222252e-04],
        [-6.68698723e-04, 2.61494786e-03, -4.68601905e-05],
        [-1.67939714e-04, 1.02402038e-05, 4.19142287e-03],
    ]
)
ref_electronicgtensor_large = np.array(
    [
        [0.30798458, -0.07215188, 0.07146523],
        [-0.07215626, 0.82904402, -0.52638538],
        [0.07145804, -0.5263783, 0.82010589],
    ]
)


def test_electronicgtensor_tiny() -> None:
    mol = molecules.molecule_lih_cation_sto3g()
    mol.build()

    mf = pyscf.scf.uhf.UHF(mol)
    mf.scf()

    assert isinstance(mf.mo_coeff, np.ndarray)
    assert isinstance(mf.mo_energy, np.ndarray)
    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    gtensor_calculator = magnetic.ElectronicGTensor(
        Program.PySCF, mol, cphf.CPHF(solvers.ExactInv(C, E, occupations)), C, E, occupations
    )
    gtensor_calculator.form_operators()
    gtensor_calculator.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    gtensor_calculator.form_results()

    print(ref_electronicgtensor_tiny)
    print(gtensor_calculator.g_oz_soc_1)

    assert np.all(
        np.equal(
            np.sign(ref_electronicgtensor_tiny),
            np.sign(utils.screen(gtensor_calculator.g_oz_soc_1)),
        )
    )


def test_electronicgtensor_small() -> None:
    mol = molecules.molecule_bc2h4_neutral_radical_sto3g()
    mol.build()

    mf = pyscf.scf.uhf.UHF(mol)
    mf.scf()

    assert isinstance(mf.mo_coeff, np.ndarray)
    assert isinstance(mf.mo_energy, np.ndarray)
    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    gtensor_calculator = magnetic.ElectronicGTensor(
        Program.PySCF, mol, cphf.CPHF(solvers.ExactInv(C, E, occupations)), C, E, occupations
    )
    gtensor_calculator.form_operators()
    gtensor_calculator.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    gtensor_calculator.form_results()

    print(ref_electronicgtensor_small)
    print(gtensor_calculator.g_oz_soc_1)

    assert np.all(
        np.equal(np.sign(ref_electronicgtensor_small), np.sign(gtensor_calculator.g_oz_soc_1))
    )


# def test_electronicgtensor_large() -> None:

#     mol = molecules.molecule_0w4a_dication_321g()
#     mol.build()

#     mf = pyscf.scf.uhf.UHF(mol)
#     mf.scf()

#     assert isinstance(mf.mo_coeff, np.ndarray)
#     assert isinstance(mf.mo_energy, np.ndarray)
#     C = utils.fix_mocoeffs_shape(mf.mo_coeff)
#     E = utils.fix_moenergies_shape(mf.mo_energy)
#     occupations = occupations_from_pyscf_mol(mol, C)

#     gtensor_calculator = magnetic.ElectronicGTensor(mol, C, E, occupations, hamiltonian='rpa', spin='singlet')
#     gtensor_calculator.form_operators()
#     gtensor_calculator.run()
#     gtensor_calculator.form_results()

#     print(ref_electronicgtensor_large)
#     print(gtensor_calculator.g_oz_soc_1)

#     assert np.all(np.equal(np.sign(ref_electronicgtensor_large),
#                            np.sign(gtensor_calculator.g_oz_soc_1)))

#     return



r"""Explicit equations for orbital Hessian terms using partially-transformed MO-basis two-electron integrals, *e.g.*, :math:`(ia|jb), (ij|ab)`."""

import numpy as np

from pymolresponse.utils import form_vec_energy_differences


def form_rpa_a_matrix_mo_singlet_partial(
    E_MO: np.ndarray, TEI_MO_iajb: np.ndarray, TEI_MO_ijab: np.ndarray
) -> np.ndarray:
    r"""Form the A (CIS) matrix in the MO basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is
    :math:`\left<aj||ib\right> = \left<aj|ib\right> -
    \left<aj|bi\right> = [ai|jb] - [ab|ji] = 2(ai|jb) - (ab|ji)`. It
    also includes the virt-occ energy difference on the diagonal.
    """

    shape_iajb = TEI_MO_iajb.shape
    shape_ijab = TEI_MO_ijab.shape
    assert len(shape_iajb) == len(shape_ijab) == 4
    assert shape_iajb[0] == shape_iajb[2] == shape_ijab[0] == shape_ijab[1]
    assert shape_iajb[1] == shape_iajb[3] == shape_ijab[2] == shape_ijab[3]
    nocc = shape_iajb[0]
    nvirt = shape_iajb[1]
    norb = nocc + nvirt
    assert len(E_MO.shape) == 2
    assert E_MO.shape[0] == E_MO.shape[1] == norb
    nov = nocc * nvirt

    ediff = form_vec_energy_differences(np.diag(E_MO)[:nocc], np.diag(E_MO)[nocc:])

    A = 2 * TEI_MO_iajb
    A -= TEI_MO_ijab.swapaxes(1, 2)
    A.shape = (nov, nov)

    A += np.diag(ediff)

    return A


def form_rpa_a_matrix_mo_triplet_partial(E_MO: np.ndarray, TEI_MO_ijab: np.ndarray) -> np.ndarray:
    r"""Form the A (CIS) matrix in the MO basis. [triplet]

    The equation for element :math:`\{ia,jb\}` is :math:`-
    \left<aj|bi\right> = - [ab|ji] = - (ab|ji)`. It also includes the
    virt-occ energy difference on the diagonal.
    """

    shape_ijab = TEI_MO_ijab.shape
    assert len(shape_ijab) == 4
    assert shape_ijab[0] == shape_ijab[1]
    assert shape_ijab[2] == shape_ijab[3]
    nocc = shape_ijab[0]
    nvirt = shape_ijab[2]
    norb = nocc + nvirt
    assert len(E_MO.shape) == 2
    assert E_MO.shape[0] == E_MO.shape[1] == norb
    nov = nocc * nvirt

    ediff = form_vec_energy_differences(np.diag(E_MO)[:nocc], np.diag(E_MO)[nocc:])

    A = np.zeros((nocc, nvirt, nocc, nvirt))
    A -= TEI_MO_ijab.swapaxes(1, 2)
    A.shape = (nov, nov)

    A += np.diag(ediff)

    return A


def form_rpa_b_matrix_mo_singlet_partial(TEI_MO_iajb: np.ndarray) -> np.ndarray:
    r"""Form the B matrix for RPA in the MO basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is
    :math:`\left<ab||ij\right> = \left<ab|ij\right> -
    \left<ab|ji\right> = [ai|bj] - [aj|bi] = 2(ai|bj) - (aj|bi)`.
    """

    shape_iajb = TEI_MO_iajb.shape
    assert len(shape_iajb) == 4
    assert shape_iajb[0] == shape_iajb[2]
    assert shape_iajb[1] == shape_iajb[3]
    nocc = shape_iajb[0]
    nvirt = shape_iajb[1]
    nov = nocc * nvirt

    B = 2 * TEI_MO_iajb
    B -= TEI_MO_iajb.swapaxes(1, 3)
    B.shape = (nov, nov)

    return -B


def form_rpa_b_matrix_mo_triplet_partial(TEI_MO_iajb: np.ndarray) -> np.ndarray:
    r"""Form the B matrix for RPA in the MO basis. [triplet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    shape_iajb = TEI_MO_iajb.shape
    assert len(shape_iajb) == 4
    assert shape_iajb[0] == shape_iajb[2]
    assert shape_iajb[1] == shape_iajb[3]
    nocc = shape_iajb[0]
    nvirt = shape_iajb[1]
    nov = nocc * nvirt

    B = np.zeros((nocc, nvirt, nocc, nvirt))
    B -= TEI_MO_iajb.swapaxes(1, 3)
    B.shape = (nov, nov)

    return -B


def form_rpa_a_matrix_mo_singlet_ss_partial(
    E_MO: np.ndarray, TEI_MO_iajb: np.ndarray, TEI_MO_ijab: np.ndarray
) -> np.ndarray:
    r"""Form the same-spin part of the A (CIS) matrix in the MO
    basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    shape_iajb = TEI_MO_iajb.shape
    shape_ijab = TEI_MO_ijab.shape
    assert len(shape_iajb) == len(shape_ijab) == 4
    assert shape_iajb[0] == shape_iajb[2] == shape_ijab[0] == shape_ijab[1]
    assert shape_iajb[1] == shape_iajb[3] == shape_ijab[2] == shape_ijab[3]
    nocc = shape_iajb[0]
    nvirt = shape_iajb[1]
    norb = nocc + nvirt
    assert len(E_MO.shape) == 2
    assert E_MO.shape[0] == E_MO.shape[1] == norb
    nov = nocc * nvirt

    ediff = form_vec_energy_differences(np.diag(E_MO)[:nocc], np.diag(E_MO)[nocc:])

    A = TEI_MO_iajb.copy()
    A -= TEI_MO_ijab.swapaxes(1, 2)
    A.shape = (nov, nov)

    A += np.diag(ediff)

    return A


def form_rpa_a_matrix_mo_singlet_os_partial(TEI_MO_iajb_xxyy: np.ndarray) -> np.ndarray:
    r"""Form the opposite-spin part of the A (CIS) matrix in the MO
    basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    shape = TEI_MO_iajb_xxyy.shape
    assert len(shape) == 4
    nocc_x, nvirt_x, nocc_y, nvirt_y = shape
    nov_x = nocc_x * nvirt_x
    nov_y = nocc_y * nvirt_y

    A = TEI_MO_iajb_xxyy.copy()
    A.shape = (nov_x, nov_y)

    return A


def form_rpa_b_matrix_mo_singlet_ss_partial(TEI_MO_iajb: np.ndarray) -> np.ndarray:
    r"""Form the same-spin part of the RPA B matrix in the MO
    basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    shape_iajb = TEI_MO_iajb.shape
    assert len(shape_iajb) == 4
    assert shape_iajb[0] == shape_iajb[2]
    assert shape_iajb[1] == shape_iajb[3]
    nocc = shape_iajb[0]
    nvirt = shape_iajb[1]
    nov = nocc * nvirt

    B = TEI_MO_iajb.copy()
    B -= TEI_MO_iajb.swapaxes(1, 3)
    B.shape = (nov, nov)

    return -B


def form_rpa_b_matrix_mo_singlet_os_partial(TEI_MO_iajb_xxyy: np.ndarray) -> np.ndarray:
    r"""Form the opposite-spin part of the RPA B matrix in the MO
    basis. [singlet]

    The equation for element :math:`\{ia,jb\}` is :math:`????`.
    """

    shape = TEI_MO_iajb_xxyy.shape
    assert len(shape) == 4
    nocc_x, nvirt_x, nocc_y, nvirt_y = shape
    nov_x = nocc_x * nvirt_x
    nov_y = nocc_y * nvirt_y

    B = TEI_MO_iajb_xxyy.copy()
    B.shape = (nov_x, nov_y)

    return -B



import numpy as np
import scipy as sp

import psi4
import pyscf

from pymolresponse import cphf, solvers, utils
from pymolresponse.core import Hamiltonian, Program, Spin
from pymolresponse.interfaces.psi4 import integrals
from pymolresponse.interfaces.psi4 import molecules as molecules_psi4
from pymolresponse.interfaces.psi4.utils import (
    mocoeffs_from_psi4wfn,
    moenergies_from_psi4wfn,
    occupations_from_psi4wfn,
)
from pymolresponse.interfaces.pyscf import molecules as molecules_pyscf
from pymolresponse.interfaces.pyscf.utils import occupations_from_pyscf_mol
from pymolresponse.properties import electric, magnetic


def test_inversion() -> None:
    """Test that each kind of inversion function gives identical results."""

    mol = molecules_pyscf.molecule_glycine_sto3g()
    mol.charge = 1
    mol.spin = 1
    mol.build()

    mf = pyscf.scf.uhf.UHF(mol)
    mf.scf()

    assert isinstance(mf.mo_coeff, np.ndarray)
    assert isinstance(mf.mo_energy, np.ndarray)
    assert len(mf.mo_coeff) == 2
    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    calculator_ref = magnetic.Magnetizability(
        Program.PySCF, mol, cphf.CPHF(solvers.ExactInv(C, E, occupations)), C, E, occupations
    )
    calculator_ref.form_operators()
    calculator_ref.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    calculator_ref.form_results()

    inv_funcs = (sp.linalg.inv, sp.linalg.pinv)

    thresh = 6.0e-14

    # TODO actually test the different inversion functions...
    for _ in inv_funcs:
        calculator_res = magnetic.Magnetizability(
            Program.PySCF, mol, cphf.CPHF(solvers.ExactInv(C, E, occupations)), C, E, occupations
        )
        calculator_res.form_operators()
        calculator_res.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
        calculator_res.form_results()

        np.testing.assert_equal(
            np.sign(calculator_ref.magnetizability), np.sign(calculator_res.magnetizability)
        )
        diff = calculator_ref.magnetizability - calculator_res.magnetizability
        abs_diff = np.abs(diff)
        print(abs_diff)
        assert np.all(abs_diff < thresh)


def test_final_result_rhf_h2o_sto3g_rpa_singlet_iter() -> None:
    mol = molecules_psi4.molecule_physicists_water_sto3g()
    psi4.core.set_active_molecule(mol)
    psi4.set_options(
        {
            "scf_type": "direct",
            "df_scf_guess": False,
            "e_convergence": 1e-11,
            "d_convergence": 1e-11,
        }
    )
    _, wfn = psi4.energy("hf", return_wfn=True)
    C = mocoeffs_from_psi4wfn(wfn)
    E = moenergies_from_psi4wfn(wfn)
    occupations = occupations_from_psi4wfn(wfn)

    frequencies = [0.0, 0.0773178]

    ref_polarizability = electric.Polarizability(
        Program.Psi4,
        mol,
        cphf.CPHF(solvers.ExactInv(C, E, occupations)),
        C,
        E,
        occupations,
        frequencies=frequencies,
    )
    ref_polarizability.form_operators()
    ref_polarizability.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    ref_polarizability.form_results()
    ref_operator = ref_polarizability.driver.solver.operators[0]  # noqa: F841
    res_polarizability = electric.Polarizability(
        Program.Psi4,
        mol,
        cphf.CPHF(
            solvers.IterativeLinEqSolver(C, E, occupations, integrals.JKPsi4(wfn), conv=1.0e-12)
        ),
        C,
        E,
        occupations,
        frequencies=frequencies,
    )
    res_polarizability.form_operators()
    res_polarizability.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    res_polarizability.form_results()
    res_operator = res_polarizability.driver.solver.operators[0]  # noqa: F841

    np.testing.assert_allclose(
        ref_polarizability.polarizabilities,
        res_polarizability.polarizabilities,
        rtol=0.0,
        atol=1.0e-6,
    )



from typing import Optional

import numpy as np
import scipy.constants as spc

from pymolresponse.utils import fix_mocoeffs_shape, repack_matrix_to_vector


class Operator:
    """Handle property integrals, taking them from the AO basis to a
    representation of a right-hand side perturbation for CPHF or
    transition properties."""

    def __init__(
        self,
        label: str = "",
        is_imaginary: bool = False,
        is_spin_dependent: bool = False,
        triplet: bool = False,
        slice_idx: int = -1,
        ao_integrals: Optional[np.ndarray] = None,
    ) -> None:
        self.label = label
        self.is_imaginary = is_imaginary
        self.is_spin_dependent = is_spin_dependent
        self.triplet = triplet
        # TODO In general, this is not used outside of referencing a
        # specific operator component from DALTON.
        self.slice_idx = slice_idx

        if "spinorb" in label:
            self.hsofac = (spc.alpha**2) / 4

        self.frequencies = None
        self.ao_integrals = ao_integrals
        self.rspvecs_alph = []
        self.rspvecs_beta = []

    def __str__(self) -> str:
        return (
            f'Operator(label="{self.label}", is_imaginary={self.is_imaginary}, '
            f"is_spin_dependent={self.is_spin_dependent}, triplet={self.triplet}, "
            f"slice_idx={self.slice_idx})"
        )

    def calculate_ao_integrals(self) -> None:
        pass

    def form_rhs(self, C: np.ndarray, occupations: np.ndarray) -> None:
        """Form the right-hand side for CPHF."""
        assert isinstance(self.ao_integrals, np.ndarray)
        if len(C.shape) == 2:
            C = C[np.newaxis]
        assert len(C.shape) == 3
        assert (C.shape[0] == 1) or (C.shape[0] == 2)
        is_uhf = C.shape[0] == 2
        C_alph = C[0]
        if is_uhf:
            C_beta = C[1]
        assert len(occupations) == 4
        nocc_alph, _, nocc_beta, _ = occupations
        b_prefactor = -1
        if self.is_imaginary:
            b_prefactor = +1
        operator_ai_alph = []
        operator_ai_supervector_alph = []
        operator_ai_beta = []
        operator_ai_supervector_beta = []
        # Loop over the operator components (usually multiple
        # Cartesian directions).
        # pylint: disable=no-member
        for idx in range(self.ao_integrals.shape[0]):
            operator_component_ai_alph = np.dot(
                C_alph[:, nocc_alph:].T, np.dot(self.ao_integrals[idx], C_alph[:, :nocc_alph])
            )
            # If the operator is a triplet operator and doing singlet
            # response, remove inactive -> secondary excitations.
            # Is this only true for spin-orbit operators?
            if self.triplet:
                for i, a in self.indices_closed_secondary:
                    operator_component_ai_alph[a - nocc_alph, i] = 0.0
            operator_component_ai_alph = repack_matrix_to_vector(operator_component_ai_alph)[
                :, np.newaxis
            ]
            if hasattr(self, "hsofac"):
                operator_component_ai_alph *= self.hsofac
            operator_component_ai_supervector_alph = np.concatenate(
                (operator_component_ai_alph, operator_component_ai_alph * b_prefactor), axis=0
            )
            operator_ai_alph.append(operator_component_ai_alph)
            operator_ai_supervector_alph.append(operator_component_ai_supervector_alph)
            if is_uhf:
                operator_component_ai_beta = np.dot(
                    C_beta[:, nocc_beta:].T, np.dot(self.ao_integrals[idx], C_beta[:, :nocc_beta])
                )
                if self.triplet:
                    for i, a in self.indices_closed_secondary:
                        operator_component_ai_beta[a - nocc_beta, i] = 0.0
                operator_component_ai_beta = repack_matrix_to_vector(operator_component_ai_beta)[
                    :, np.newaxis
                ]
                if hasattr(self, "hsofac"):
                    operator_component_ai_beta *= self.hsofac
                operator_component_ai_supervector_beta = np.concatenate(
                    (operator_component_ai_beta, operator_component_ai_beta * b_prefactor), axis=0
                )
                operator_ai_beta.append(operator_component_ai_beta)
                operator_ai_supervector_beta.append(operator_component_ai_supervector_beta)
        self.mo_integrals_ai_alph = np.stack(operator_ai_alph, axis=0)
        self.mo_integrals_ai_supervector_alph = np.stack(operator_ai_supervector_alph, axis=0)
        if is_uhf:
            self.mo_integrals_ai_beta = np.stack(operator_ai_beta, axis=0)
            self.mo_integrals_ai_supervector_beta = np.stack(operator_ai_supervector_beta, axis=0)

    def form_rhs_geometric(
        self,
        C: np.ndarray,
        occupations: np.ndarray,
        natoms,
        MO_full,
        mints,
        return_dict: bool = False,
    ) -> None:
        from pymolresponse.integrals import _form_rhs_geometric

        C_ = fix_mocoeffs_shape(C)
        B_dict = _form_rhs_geometric(C_[0], occupations, natoms, MO_full, mints)
        if return_dict:
            return B_dict
        # TODO I think this is wrong, because the matrices are [i, a]
        B_matrices = [B_dict[k].T for k in sorted(B_dict.keys())]
        B_vectors = [repack_matrix_to_vector(B)[:, np.newaxis] for B in B_matrices]
        mo_integrals_ai_alph = np.stack(B_vectors)
        self.mo_integrals_ai_alph = mo_integrals_ai_alph
        # pylint: disable=invalid-unary-operand-type
        self.mo_integrals_ai_supervector_alph = np.concatenate(
            (mo_integrals_ai_alph, -mo_integrals_ai_alph), axis=1
        )



"""Utility functions that are not core to calculating physical values."""

from itertools import accumulate
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np


def form_results(vecs_property: np.ndarray, vecs_response: np.ndarray) -> np.ndarray:
    assert vecs_property.shape[1:] == vecs_response.shape[1:]
    assert len(vecs_property.shape) == 3
    assert vecs_property.shape[2] == 1
    results = np.dot(vecs_property[:, :, 0], vecs_response[:, :, 0].T)
    return results


def np_load(filename: Union[str, Path]) -> np.ndarray:
    """Read a file using NumPy."""
    arr = np.load(filename)
    if isinstance(arr, np.lib.npyio.NpzFile):
        # Make the assumption that there's only a single array
        # present, even though *.npz files can hold multiple arrays.
        for _arr_name, _arr in arr.items():
            arr = _arr
            break
    return arr


def parse_int_file_2(
    filename: Union[str, Path], dim: int
) -> np.ndarray[Tuple[int, int], np.dtype[float]]:
    mat = np.zeros(shape=(dim, dim))
    with open(filename) as fh:
        contents = fh.readlines()
    for line in contents:
        mu, nu, intval = [float(x) for x in line.split()]
        mu, nu = int(mu - 1), int(nu - 1)
        mat[mu, nu] = mat[nu, mu] = intval
    return mat


def repack_matrix_to_vector(mat: np.ndarray) -> np.ndarray:
    return np.reshape(mat, -1, order="F")


def repack_vector_to_matrix(vec: np.ndarray, shape: Tuple[int, ...]) -> np.ndarray:
    return vec.reshape(shape, order="F")


def get_reference_value_from_file(
    filename: Union[Path, str],
    hamiltonian: str,
    spin: str,
    frequency: str,
    label_1: str,
    label_2: str,
) -> float:
    # TODO need to pass the frequency as a string identical to the one
    # found in the file, can't pass a float due to fp error; how to
    # get around this?
    found = False
    with open(filename) as fh:
        for line in fh:
            tokens = line.split()
            # no comments allowed for now
            assert len(tokens) == 6
            l_hamiltonian, l_spin, l_frequency, l_label_1, l_label_2, l_val = tokens
            if (
                (l_hamiltonian == hamiltonian)
                and (l_spin == spin)
                and (l_frequency == frequency)
                and (l_label_1 == label_1)
                and (l_label_2 == label_2)
            ):
                ref = float(l_val)
                found = True

    if not found:
        raise ValueError("Could not find reference value")

    return ref


def read_file_occupations(filename: Union[Path, str]) -> np.ndarray:
    with open(filename) as fh:
        contents = fh.read().strip()
    tokens = contents.split()
    assert len(tokens) == 4
    nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = [int(x) for x in tokens]
    return np.asarray([nocc_alph, nvirt_alph, nocc_beta, nvirt_beta], dtype=int)


def read_file_1(filename: Union[Path, str]) -> np.ndarray:
    elements = []
    with open(filename) as fh:
        n_elem = int(next(fh))
        for line in fh:
            elements.append(float(line))
    assert len(elements) == n_elem
    return np.array(elements, dtype=float)


def read_file_2(filename: Union[Path, str]) -> np.ndarray:
    elements = []
    with open(filename) as fh:
        n_rows, n_cols = [int(x) for x in next(fh).split()]
        for line in fh:
            elements.append(float(line))
    assert len(elements) == (n_rows * n_cols)
    # The last index is the fast index (cols).
    return np.reshape(np.array(elements, dtype=float), (n_rows, n_cols))


def read_file_3(filename: Union[Path, str]) -> np.ndarray:
    elements = []
    with open(filename) as fh:
        n_slices, n_rows, n_cols = [int(x) for x in next(fh).split()]
        for line in fh:
            elements.append(float(line))
    assert len(elements) == (n_rows * n_cols * n_slices)
    return np.reshape(np.array(elements, dtype=float), (n_slices, n_rows, n_cols))


def read_file_4(filename: Union[Path, str]) -> np.ndarray:
    elements = []
    with open(filename) as fh:
        n_d1, n_d2, n_d3, n_d4 = [int(x) for x in next(fh).split()]
        for line in fh:
            elements.append(float(line))
    assert len(elements) == (n_d1 * n_d2 * n_d3 * n_d4)
    return np.reshape(np.array(elements, dtype=float), (n_d1, n_d2, n_d3, n_d4))


def occupations_from_sirifc(ifc) -> np.ndarray:
    nocc_a, nocc_b = ifc.nisht + ifc.nasht, ifc.nisht
    norb = ifc.norbt
    nvirt_a, nvirt_b = norb - nocc_a, norb - nocc_b
    return np.asarray([nocc_a, nvirt_a, nocc_b, nvirt_b], dtype=int)


class Splitter:
    """Split a line based not on a character, but a given number of field
    widths.
    """

    def __init__(self, widths) -> None:
        self.start_indices = [0] + list(accumulate(widths))[:-1]
        self.end_indices = list(accumulate(widths))

    def split(self, line: str, truncate: bool = True) -> List[str]:
        """Split the given line using the field widths passed in on class
        initialization.

        Handle lines that contain fewer fields than specified in the
        widths; they are added as empty strings. If `truncate`, remove
        them.
        """
        elements = [
            line[start:end].strip() for (start, end) in zip(self.start_indices, self.end_indices)
        ]
        if truncate:
            for _ in range(1, len(elements)):
                if elements[-1] == "":
                    elements.pop()
                else:
                    break
        return elements


def fix_mocoeffs_shape(
    mocoeffs: Union[
        Tuple[np.ndarray[Tuple[int, ...], np.dtype[np.floating]], ...],
        np.ndarray[Union[Tuple[int, int], Tuple[int, int, int]], np.dtype[np.floating]],
    ],
) -> np.ndarray[Tuple[int, int, int], np.dtype[np.floating]]:
    if isinstance(mocoeffs, tuple):
        # this will properly fall through to the else clause
        mocoeffs_new = fix_mocoeffs_shape(np.stack(mocoeffs, axis=0))
    # assume np.ndarray
    else:
        shape = mocoeffs.shape
        assert len(shape) in (2, 3)
        if len(shape) == 2:
            mocoeffs_new = mocoeffs[np.newaxis]
        else:
            mocoeffs_new = mocoeffs
    assert len(mocoeffs_new.shape) == 3
    return mocoeffs_new  # ty: ignore[invalid-return-type]


def fix_moenergies_shape(
    moenergies: Union[
        Tuple[np.ndarray[Tuple[int, ...], np.dtype[np.floating]], ...],
        np.ndarray[Union[Tuple[int], Tuple[int, int], Tuple[int, int, int]], np.dtype[np.floating]],
    ],
) -> np.ndarray[Tuple[int, int, int], np.dtype[np.floating]]:
    if isinstance(moenergies, tuple):
        # this will properly fall through to the else clause
        moenergies_new = fix_moenergies_shape(np.stack(moenergies, axis=0))
    else:
        shape = moenergies.shape
        ls = len(shape)
        assert ls in (1, 2, 3)
        if ls == 1:
            # It's a vector.
            moenergies_new = np.diag(moenergies)[np.newaxis]
        elif ls == 2:
            # If it's a square matrix, assume it's already diagonal.  If it
            # isn't a square matrix, then it probably has one or two columns,
            # one for each spin case.  TODO check that all off-diagonal
            # elements are zero?  Not true for Fock matrix in non-orthogonal
            # basis.
            if shape[0] == shape[1]:  # ty: ignore[index-out-of-bounds]
                moenergies_new = moenergies[np.newaxis]
            else:
                assert shape[0] in (1, 2)
                if shape[0] == 1:
                    # (1, norb)
                    # FIXME swapped?
                    moenergies_new = np.diag(moenergies[:, 0])[np.newaxis]
                else:
                    # (2, norb)
                    moenergies_alph = np.diag(moenergies[0, :])[np.newaxis]
                    moenergies_beta = np.diag(moenergies[1, :])[np.newaxis]
                    moenergies_new = np.concatenate((moenergies_alph, moenergies_beta), axis=0)
        else:
            assert shape[0] in (1, 2)
            # You might think at first glance there's an assumption that nbsf
            # == nmo here, but the (Fock) matrix is entirely in the MO basis.
            assert shape[1] == shape[2]  # ty: ignore[index-out-of-bounds]
            moenergies_new = moenergies
    assert len(moenergies_new.shape) == 3
    assert moenergies_new.shape[0] in (1, 2)
    assert moenergies_new.shape[1] == moenergies_new.shape[2]  # ty: ignore[index-out-of-bounds]
    return moenergies_new  # ty: ignore[invalid-return-type]


def read_dalton_propfile(tmpdir: Path) -> List[str]:
    proplist = []
    with open(tmpdir / "DALTON.PROP") as propfile:
        proplines = propfile.readlines()
    splitter = Splitter([5, 3, 4, 11, 23, 9, 9, 9, 9, 23, 23, 23, 4, 4, 4])
    for line in proplines:
        sline = splitter.split(line)
        # print(sline)
        proplist.append(sline)
    return proplist


def tensor_printer(tensor: np.ndarray) -> Tuple[np.ndarray, float, float]:
    print(tensor)
    eigvals = np.linalg.eigvals(tensor)
    # or should this be the trace of the matrix?
    iso = np.average(eigvals)
    # anisotropic = 0.e0;
    # for(int j1=0;j1 < 9; j1++)
    #   anisotropic += statpol[j1]*statpol[j1];
    # anisotropic = sqrt(fabs(1.5e0 * (anisotropic - 3*isotropic * isotropic)));
    # aniso = np.sum(tensor ** 2)
    aniso = 0.0
    print(eigvals)
    print(iso)
    # print(np.trace(tensor) / tensor.shape[0])
    # print(aniso)
    return eigvals, iso, aniso


def form_vec_energy_differences(moene_occ: np.ndarray, moene_virt: np.ndarray) -> np.ndarray:
    nocc = moene_occ.shape[0]
    nvirt = moene_virt.shape[0]
    nov = nocc * nvirt
    # The stupid loop is faster!
    ediff = np.zeros(nov)
    for i in range(nocc):
        for a in range(nvirt):
            ia = (i * nvirt) + a
            ediff[ia] = moene_virt[a] - moene_occ[i]
    # eo = np.einsum('ij,ab->iajb', np.diag(moene_occ), np.diag(np.ones(nvirt)))
    # ev = np.einsum('ab,ij->iajb', np.diag(moene_virt), np.diag(np.ones(nocc)))
    # ediff = np.diag((ev - eo).reshape(nov, nov))
    return ediff


def screen(mat: np.ndarray, thresh: float = 1.0e-16) -> np.ndarray:
    """Set all values smaller than the given threshold to zero
    (considering them as numerical noise).

    Parameters
    ----------
    mat : np.ndarray
    thresh : float
        Threshold below which all elements of `mat` smaller than
        `thresh` are set to zero.

    Returns
    -------
    np.ndarray
    """
    mat_screened = mat.copy()
    mat_screened[np.abs(mat) <= thresh] = 0.0
    return mat_screened


def matsym(amat: np.ndarray, thrzer: float = 1.0e-14) -> int:
    """
    - Copied from ``DALTON/gp/gphjj.F/MATSYM``.
    - `thrzer` taken from ``DALTON/include/thrzer.h``.

    Parameters
    ----------
    amat : np.ndarray
    thrzer : float
        Threshold below which a (floating point) number is considered
        zero.

    Returns
    -------
    int
        - 1 if the matrix is symmetric to threshold `thrzer`
        - 2 if the matrix is antisymmetric to threshold `thrzer`
        - 3 if all elements are below `thrzer`
        - 0 otherwise (the matrix is unsymmetric about the diagonal)
    """

    assert amat.shape[0] == amat.shape[1]

    n = amat.shape[0]

    isym = 1
    iasym = 2
    for j in range(n):
        # for i in range(j+1):
        # The +1 is so the diagonal elements are checked.
        for i in range(j + 1):
            amats = abs(amat[i, j] + amat[j, i])
            amata = abs(amat[i, j] - amat[j, i])
            if amats > thrzer:
                iasym = 0
            if amata > thrzer:
                isym = 0

    return isym + iasym


def flip_triangle_sign(A: np.ndarray, triangle: str = "lower") -> np.ndarray:
    """Flip the sign of either the lower or upper triangle of a sqare
    matrix. Assume nothing about its symmetry.

    Parameters
    ----------
    A : np.ndarray
    triangle : {'lower', 'upper'}

    Returns
    -------
    np.ndarray
    """
    assert len(A.shape) == 2
    assert A.shape[0] == A.shape[1]
    dim = A.shape[0]
    if triangle == "lower":
        indices = np.tril_indices(dim)
    elif triangle == "upper":
        indices = np.triu_indices(dim)
    else:
        raise ValueError("argument to triangle must be 'upper' or 'lower'")
    B = A.copy()
    B[indices] *= -1.0
    return B


def form_first_hyperpolarizability_averages(beta: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    assert beta.shape == (3, 3, 3)
    avgs = (-1 / 3) * (
        np.einsum("ijj->i", beta) + np.einsum("jij->i", beta) + np.einsum("jji->i", beta)
    )
    avg = np.sum(avgs**2) ** (1 / 2)
    return avgs, avg


def form_indices_orbwin(nocc: int, nvirt: int) -> List[Tuple[int, int]]:
    norb = nocc + nvirt
    return [(i, a) for i in range(0, nocc) for a in range(nocc, norb)]


def form_indices_zero(nocc: int, nvirt: int) -> List[Tuple[int, int]]:
    return [(i, a) for i in range(nocc) for a in range(nvirt)]



import numpy as np
import pytest

from pymolresponse import molecular_property
from pymolresponse.core import Program


@pytest.mark.skip("TODO these have been turned into ABCs")
def test_molecular_property() -> None:
    pyscfmol = None
    mocoeffs = np.array([])
    moenergies = np.array([])
    occupations = np.array([])
    cls = molecular_property.MolecularProperty(
        Program.PySCF, pyscfmol, mocoeffs, moenergies, occupations
    )
    try:
        cls.form_operators()
    except NotImplementedError as e:
        message = e.args[0]
        assert message == "This must be implemented in a grandchild class."
    try:
        cls.form_results()
    except NotImplementedError as e:
        message = e.args[0]
        assert message == "This must be implemented in a grandchild class."


@pytest.mark.skip()
def test_response_property() -> None:
    pyscfmol = None  # noqa: F841
    mocoeffs = np.zeros((1, 2, 2))  # noqa: F841
    moenergies = np.zeros((1, 2, 2))  # noqa: F841
    occupations = [0 for _ in range(4)]  # noqa: F841
    # TODO Turns out invoking the solver does a lot automatically...
    # cls = molecular_property.ResponseProperty(pyscfmol, mocoeffs, moenergies, occupations)
    # try:
    #     cls.form_operators()
    # except NotImplementedError as e:
    #     message = e.args[0]
    #     assert message == "This must be implemented in a child class."
    # try:
    #     cls.form_results()
    # except NotImplementedError as e:
    #     message = e.args[0]
    #     assert message == "This must be implemented in a child class."


@pytest.mark.skip()
def test_transition_property() -> None:
    pyscfmol = None  # noqa: F841
    mocoeffs = np.zeros((1, 2, 2))  # noqa: F841
    moenergies = np.zeros((1, 2, 2))  # noqa: F841
    occupations = [0 for _ in range(4)]  # noqa: F841
    # TODO Turns out invoking the solver does a lot automatically...
    # cls = molecular_property.TransitionProperty(pyscfmol, mocoeffs, moenergies, occupations)
    # try:
    #     cls.form_operators()
    # except NotImplementedError as e:
    #     message = e.args[0]
    #     assert message == "This must be implemented in a child class."
    # try:
    #     cls.form_results()
    # except NotImplementedError as e:
    #     message = e.args[0]
    #     assert message == "This must be implemented in a child class."



import numpy as np

import psi4

from pymolresponse.interfaces.psi4 import integrals, molecules
from pymolresponse.interfaces.psi4.utils import mocoeffs_from_psi4wfn, occupations_from_psi4wfn


def test_integrals_psi4() -> None:
    mol = molecules.molecule_water_sto3g()
    mol.reset_point_group("c1")
    mol.update_geometry()
    psi4.core.set_active_molecule(mol)

    options = {"BASIS": "STO-3G", "SCF_TYPE": "PK", "E_CONVERGENCE": 1e-10, "D_CONVERGENCE": 1e-10}

    psi4.set_options(options)

    _, wfn = psi4.energy("hf", return_wfn=True)

    mints = psi4.core.MintsHelper(wfn)

    integral_generator = integrals.IntegralsPsi4(wfn)

    np.testing.assert_equal(
        np.stack([np.asarray(Mc) for Mc in mints.ao_dipole()]),
        integral_generator.integrals(integrals.DIPOLE),
    )


def test_jk_psi4() -> None:
    mol = molecules.molecule_water_sto3g()
    mol.reset_point_group("c1")
    mol.update_geometry()
    psi4.core.set_active_molecule(mol)

    options = {"BASIS": "STO-3G", "SCF_TYPE": "PK", "E_CONVERGENCE": 1e-10, "D_CONVERGENCE": 1e-10}

    psi4.set_options(options)

    _, wfn = psi4.energy("hf", return_wfn=True)
    C = mocoeffs_from_psi4wfn(wfn)
    jk_generator = integrals.JKPsi4(wfn)
    res_J, res_K = jk_generator.compute_from_mocoeffs(C[0])

    # TODO comparison against reference
    occupations = occupations_from_psi4wfn(wfn)
    # TODO this assumes that the total number of orbitals is equal to the
    # total number of basis functions
    nbasis = occupations[0] + occupations[1]
    assert res_J.shape == (nbasis, nbasis)
    assert res_K.shape == (nbasis, nbasis)



from pathlib import Path

import numpy as np
from cclib.io import ccopen
from cclib.parser.utils import convertor

import pyscf

from pymolresponse import cphf, operators, solvers, utils
from pymolresponse.core import AO2MOTransformationType, Hamiltonian, Spin
from pymolresponse.interfaces.dalton.utils import dalton_label_to_operator
from pymolresponse.interfaces.pyscf.ao2mo import AO2MOpyscf

try:
    from daltools import mol as dalmol
    from daltools import sirifc
except ImportError:
    pass


def calculate_disk_rhf(
    testcasedir: Path, hamiltonian: str, spin: str, frequency: str, label_1: str, label_2: str
) -> float:
    occupations = utils.read_file_occupations(testcasedir / "occupations")
    nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = occupations
    assert nocc_alph == nocc_beta
    assert nvirt_alph == nvirt_beta
    norb = nocc_alph + nvirt_alph
    C = utils.read_file_3(testcasedir / "C")
    assert C.shape[0] == 1
    assert C.shape[2] == norb
    nbasis = C.shape[1]
    moene = utils.read_file_2(testcasedir / "moene")
    assert moene.shape == (norb, 1)
    moints_iajb_aaaa = utils.read_file_4(testcasedir / "moints_iajb_aaaa")
    moints_ijab_aaaa = utils.read_file_4(testcasedir / "moints_ijab_aaaa")
    assert moints_iajb_aaaa.shape == (nocc_alph, nvirt_alph, nocc_alph, nvirt_alph)
    assert moints_ijab_aaaa.shape == (nocc_alph, nocc_alph, nvirt_alph, nvirt_alph)

    operator_1 = dalton_label_to_operator(label_1)
    operator_2 = dalton_label_to_operator(label_2)

    operator_1_integrals_mn = utils.read_file_3(testcasedir / f"operator_mn_{operator_1.label}")
    operator_2_integrals_mn = utils.read_file_3(testcasedir / f"operator_mn_{operator_2.label}")
    # The first dimension can"t be checked since there may be multiple
    # components.
    assert operator_1_integrals_mn.shape[1:] == (nbasis, nbasis)
    assert operator_2_integrals_mn.shape[1:] == (nbasis, nbasis)

    # Only take the component/slice from the integral as determined
    # from the DALTON operator label.
    operator_1_integrals_mn = operator_1_integrals_mn[operator_1.slice_idx]
    operator_2_integrals_mn = operator_2_integrals_mn[operator_2.slice_idx]
    # However, this eliminates an axis, which needs to be added back.
    operator_1_integrals_mn = operator_1_integrals_mn[np.newaxis, ...]
    operator_2_integrals_mn = operator_2_integrals_mn[np.newaxis, ...]

    operator_1.ao_integrals = operator_1_integrals_mn
    operator_2.ao_integrals = operator_2_integrals_mn

    moene = np.diag(moene[:, 0])[np.newaxis, ...]
    assert moene.shape == (1, norb, norb)

    solver = solvers.ExactInv(C, moene, occupations)
    solver.tei_mo = (moints_iajb_aaaa, moints_ijab_aaaa)
    solver.tei_mo_type = AO2MOTransformationType.partial

    driver = cphf.CPHF(solver)
    driver.add_operator(operator_1)
    driver.add_operator(operator_2)

    driver.set_frequencies([float(frequency)])

    driver.run(
        hamiltonian=Hamiltonian[hamiltonian.upper()],
        spin=Spin[spin],
        program=None,
        program_obj=None,
    )

    assert len(driver.frequencies) == len(driver.results) == 1
    res = driver.results[0]
    assert res.shape == (2, 2)
    bl = res[1, 0]
    tr = res[0, 1]
    diff = abs(abs(bl) - abs(tr))
    # Results should be symmetric w.r.t. interchange between operators
    # in the LR equations.
    thresh = 1.0e-13
    assert diff < thresh

    return bl


def calculate_disk_uhf(
    testcasedir: Path, hamiltonian: str, spin: str, frequency: str, label_1: str, label_2: str
) -> float:
    occupations = utils.read_file_occupations(testcasedir / "occupations")
    nocc_alph, nvirt_alph, nocc_beta, nvirt_beta = occupations
    norb = nocc_alph + nvirt_alph
    C = utils.read_file_3(testcasedir / "C")
    assert C.shape[0] == 2
    assert C.shape[2] == norb
    nbasis = C.shape[1]
    moene = utils.read_file_2(testcasedir / "moene")
    assert moene.shape == (norb, 2)
    moints_iajb_aaaa = utils.read_file_4(testcasedir / "moints_iajb_aaaa")
    moints_iajb_aabb = utils.read_file_4(testcasedir / "moints_iajb_aabb")
    moints_iajb_bbaa = utils.read_file_4(testcasedir / "moints_iajb_bbaa")
    moints_iajb_bbbb = utils.read_file_4(testcasedir / "moints_iajb_bbbb")
    moints_ijab_aaaa = utils.read_file_4(testcasedir / "moints_ijab_aaaa")
    moints_ijab_bbbb = utils.read_file_4(testcasedir / "moints_ijab_bbbb")
    assert moints_iajb_aaaa.shape == (nocc_alph, nvirt_alph, nocc_alph, nvirt_alph)
    assert moints_iajb_aabb.shape == (nocc_alph, nvirt_alph, nocc_beta, nvirt_beta)
    assert moints_iajb_bbaa.shape == (nocc_beta, nvirt_beta, nocc_alph, nvirt_alph)
    assert moints_iajb_bbbb.shape == (nocc_beta, nvirt_beta, nocc_beta, nvirt_beta)
    assert moints_ijab_aaaa.shape == (nocc_alph, nocc_alph, nvirt_alph, nvirt_alph)
    assert moints_ijab_bbbb.shape == (nocc_beta, nocc_beta, nvirt_beta, nvirt_beta)

    operator_1 = dalton_label_to_operator(label_1)
    operator_2 = dalton_label_to_operator(label_2)

    operator_1_integrals_mn = utils.read_file_3(testcasedir / f"operator_mn_{operator_1.label}")
    operator_2_integrals_mn = utils.read_file_3(testcasedir / f"operator_mn_{operator_2.label}")
    # The first dimension can"t be checked since there may be multiple
    # components.
    assert operator_1_integrals_mn.shape[1:] == (nbasis, nbasis)
    assert operator_2_integrals_mn.shape[1:] == (nbasis, nbasis)

    # Only take the component/slice from the integral as determined
    # from the DALTON operator label.
    operator_1_integrals_mn = operator_1_integrals_mn[operator_1.slice_idx]
    operator_2_integrals_mn = operator_2_integrals_mn[operator_2.slice_idx]
    # However, this eliminates an axis, which needs to be added back.
    operator_1_integrals_mn = operator_1_integrals_mn[np.newaxis, ...]
    operator_2_integrals_mn = operator_2_integrals_mn[np.newaxis, ...]

    operator_1.ao_integrals = operator_1_integrals_mn
    operator_2.ao_integrals = operator_2_integrals_mn

    moene_alph = np.diag(moene[:, 0])
    moene_beta = np.diag(moene[:, 1])
    moene = np.stack((moene_alph, moene_beta), axis=0)
    assert moene.shape == (2, norb, norb)

    solver = solvers.ExactInv(C, moene, occupations)
    solver.tei_mo = (
        moints_iajb_aaaa,
        moints_iajb_aabb,
        moints_iajb_bbaa,
        moints_iajb_bbbb,
        moints_ijab_aaaa,
        moints_ijab_bbbb,
    )
    solver.tei_mo_type = AO2MOTransformationType.partial

    driver = cphf.CPHF(solver)
    driver.add_operator(operator_1)
    driver.add_operator(operator_2)

    driver.set_frequencies([float(frequency)])

    driver.run(
        hamiltonian=Hamiltonian[hamiltonian.upper()],
        spin=Spin[spin],
        program=None,
        program_obj=None,
    )

    assert len(driver.frequencies) == len(driver.results) == 1
    res = driver.results[0]
    assert res.shape == (2, 2)
    bl = res[1, 0]
    tr = res[0, 1]
    diff = abs(abs(bl) - abs(tr))
    # Results should be symmetric w.r.t. interchange between operators
    # in the LR equations.
    thresh = 1.0e-14
    assert diff < thresh

    return bl


def calculate_rhf(
    dalton_tmpdir: Path,
    hamiltonian: str,
    spin: str,
    operator_label: str,
    operator: str,
    source_moenergies: str,
    source_mocoeffs: str,
    source_operator: str,
) -> float:
    if operator_label:
        # TODO add dipvel
        assert operator_label in ("dipole", "angmom", "spinorb")
    assert source_moenergies in ("pyscf", "dalton")
    assert source_mocoeffs in ("pyscf", "dalton")

    dalton_molecule = dalmol.readin(dalton_tmpdir / "DALTON.BAS")
    lines = []
    for atom in dalton_molecule:
        label = atom["label"][0]
        center = atom["center"][0]
        center_str = " ".join(["{:f}".format(pos) for pos in center])
        line = "{:3} {}".format(label, center_str)
        lines.append(line)
    lines = "\n".join(lines)

    # PySCF molecule setup, needed for generating the TEIs in the MO
    # basis.
    mol = pyscf.gto.Mole()
    verbose = 1
    mol.verbose = verbose
    mol.atom = lines
    mol.unit = "Bohr"
    # TODO read basis from DALTON molecule
    mol.basis = "sto-3g"
    mol.symmetry = False
    # TODO read charge from DALTON molecule?
    mol.charge = 0
    # TODO read spin from DALTON molecule?
    mol.spin = 0

    mol.build()

    ifc = sirifc.sirifc(dalton_tmpdir / "SIRIFC")
    occupations = utils.occupations_from_sirifc(ifc)

    if source_moenergies == "pyscf" or source_mocoeffs == "pyscf":
        mf = pyscf.scf.RHF(mol)
        mf.kernel()

    if source_moenergies == "pyscf":
        E = np.diag(mf.mo_energy)[np.newaxis, ...]
    elif source_moenergies == "dalton":
        job = ccopen(dalton_tmpdir / "DALTON.OUT")
        data = job.parse()
        # pylint: disable=no-member
        E = np.diag([convertor(x, "eV", "hartree") for x in data.moenergies[0]])[np.newaxis, ...]
    else:
        pass

    if source_mocoeffs == "pyscf":
        C = mf.mo_coeff[np.newaxis, ...]
    elif source_mocoeffs == "dalton":
        C = ifc.cmo[0][np.newaxis, ...]
    else:
        pass

    solver = solvers.ExactInv(C, E, occupations)

    solver.tei_mo = AO2MOpyscf(mol, C).perform_rhf_partial()
    solver.tei_mo_type = AO2MOTransformationType.partial

    driver = cphf.CPHF(solver)

    if operator:
        driver.add_operator(operator)
    elif operator_label:
        if operator_label == "dipole":
            operator_dipole = operators.Operator(
                label="dipole", is_imaginary=False, is_spin_dependent=False, triplet=False
            )
            integrals_dipole_ao = mol.intor("cint1e_r_sph", comp=3)
            operator_dipole.ao_integrals = integrals_dipole_ao
            driver.add_operator(operator_dipole)
        elif operator_label == "angmom":
            operator_angmom = operators.Operator(
                label="angmom", is_imaginary=True, is_spin_dependent=False, triplet=False
            )
            integrals_angmom_ao = mol.intor("cint1e_cg_irxp_sph", comp=3)
            operator_angmom.ao_integrals = integrals_angmom_ao
            driver.add_operator(operator_angmom)
        elif operator_label == "spinorb":
            operator_spinorb = operators.Operator(
                label="spinorb", is_imaginary=True, is_spin_dependent=False, triplet=False
            )
            integrals_spinorb_ao = 0
            for atm_id in range(mol.natm):
                mol.set_rinv_orig(mol.atom_coord(atm_id))
                chg = mol.atom_charge(atm_id)
                integrals_spinorb_ao += chg * mol.intor("cint1e_prinvxp_sph", comp=3)
            operator_spinorb.ao_integrals = integrals_spinorb_ao
            driver.add_operator(operator_spinorb)
        else:
            pass
    else:
        pass

    driver.set_frequencies()

    driver.run(hamiltonian=Hamiltonian[hamiltonian.upper()], spin=Spin[spin])

    return driver.results[0]


def calculate_uhf(
    dalton_tmpdir: Path,
    hamiltonian: str,
    spin: str,
    operator_label: str,
    operator: str,
    source_moenergies: str,
    source_mocoeffs: str,
    source_operator: str,
) -> float:
    if operator_label:
        # TODO add dipvel
        assert operator_label in ("dipole", "angmom", "spinorb")
    assert source_moenergies in ("pyscf", "dalton")
    assert source_mocoeffs in ("pyscf", "dalton")

    dalton_molecule = dalmol.readin(dalton_tmpdir / "DALTON.BAS")
    lines = []
    for atom in dalton_molecule:
        label = atom["label"][0]
        center = atom["center"][0]
        center_str = " ".join(["{:f}".format(pos) for pos in center])
        line = "{:3} {}".format(label, center_str)
        lines.append(line)
    lines = "\n".join(lines)

    # PySCF molecule setup, needed for generating the TEIs in the MO
    # basis.
    mol = pyscf.gto.Mole()
    verbose = 1
    mol.verbose = verbose
    mol.atom = lines
    mol.unit = "Bohr"
    # TODO read basis from DALTON molecule
    mol.basis = "sto-3g"
    mol.symmetry = False
    # TODO read charge from DALTON molecule?
    mol.charge = 1
    # TODO read spin from DALTON molecule?
    mol.spin = 1

    mol.build()

    ifc = sirifc.sirifc(dalton_tmpdir / "SIRIFC")
    occupations = utils.occupations_from_sirifc(ifc)

    if source_moenergies == "pyscf" or source_mocoeffs == "pyscf":
        mf = pyscf.scf.UHF(mol)
        mf.kernel()

    if source_moenergies == "pyscf":
        E_alph = np.diag(mf.mo_energy[0])
        E_beta = np.diag(mf.mo_energy[1])
        E = np.stack((E_alph, E_beta), axis=0)
    elif source_moenergies == "dalton":
        job = ccopen(dalton_tmpdir / "DALTON.OUT")
        data = job.parse()
        # pylint: disable=no-member
        E = np.diag([convertor(x, "eV", "hartree") for x in data.moenergies[0]])[np.newaxis, ...]
        E = np.concatenate((E, E), axis=0)
    else:
        pass

    if source_mocoeffs == "pyscf":
        C = mf.mo_coeff
    elif source_mocoeffs == "dalton":
        C = ifc.cmo[0][np.newaxis, ...]
        C = np.concatenate((C, C), axis=0)
    else:
        pass

    solver = solvers.ExactInv(C, E, occupations)

    solver.tei_mo = AO2MOpyscf(mol, C).perform_uhf_partial()
    solver.tei_mo_type = AO2MOTransformationType.partial

    driver = cphf.CPHF(solver)

    if operator:
        driver.add_operator(operator)
    elif operator_label:
        if operator_label == "dipole":
            operator_dipole = operators.Operator(
                label="dipole", is_imaginary=False, is_spin_dependent=False, triplet=False
            )
            integrals_dipole_ao = mol.intor("cint1e_r_sph", comp=3)
            operator_dipole.ao_integrals = integrals_dipole_ao
            driver.add_operator(operator_dipole)
        elif operator_label == "angmom":
            operator_angmom = operators.Operator(
                label="angmom", is_imaginary=True, is_spin_dependent=False, triplet=False
            )
            integrals_angmom_ao = mol.intor("cint1e_cg_irxp_sph", comp=3)
            operator_angmom.ao_integrals = integrals_angmom_ao
            driver.add_operator(operator_angmom)
        elif operator_label == "spinorb":
            operator_spinorb = operators.Operator(
                label="spinorb", is_imaginary=True, is_spin_dependent=False, triplet=False
            )
            integrals_spinorb_ao = 0
            for atm_id in range(mol.natm):
                mol.set_rinv_orig(mol.atom_coord(atm_id))
                chg = mol.atom_charge(atm_id)
                integrals_spinorb_ao += chg * mol.intor("cint1e_prinvxp_sph", comp=3)
            operator_spinorb.ao_integrals = integrals_spinorb_ao
            driver.add_operator(operator_spinorb)
        else:
            pass
    else:
        pass

    driver.set_frequencies()

    driver.run(hamiltonian=Hamiltonian[hamiltonian.upper()], spin=Spin[spin])

    return driver.results[0]



import numpy as np

import psi4

from pymolresponse.utils import fix_mocoeffs_shape, fix_moenergies_shape


def occupations_from_psi4wfn(wfn: psi4.core.Wavefunction) -> np.ndarray:
    # Not needed.
    # occupations_a = wfn.occupation_a().to_array()
    # occupations_b = wfn.occupation_b().to_brray()
    # assert occupations_a.shape == occupations_b.shape
    norb = wfn.nmo()
    nocc_a = wfn.nalpha()
    nocc_b = wfn.nbeta()
    nvirt_a = norb - nocc_a
    nvirt_b = norb - nocc_b
    return np.asarray([nocc_a, nvirt_a, nocc_b, nvirt_b], dtype=int)


def mocoeffs_from_psi4wfn(wfn: psi4.core.Wavefunction) -> np.ndarray:
    is_uhf = not wfn.same_a_b_orbs()
    Ca = wfn.Ca().to_array()
    if is_uhf:
        Cb = wfn.Cb().to_array()
        C = np.stack((Ca, Cb), axis=0)
    else:
        C = Ca
    # Clean up.
    return fix_mocoeffs_shape(C)


def moenergies_from_psi4wfn(wfn: psi4.core.Wavefunction) -> np.ndarray:
    is_uhf = not wfn.same_a_b_orbs()
    Ea = wfn.epsilon_a().to_array()
    if is_uhf:
        Eb = wfn.epsilon_b().to_array()
        E = np.stack((Ea, Eb), axis=0).T
    else:
        E = Ea
    # Clean up.
    return fix_moenergies_shape(E)



import numpy as np

import pyscf

from pymolresponse import cphf, solvers, utils
from pymolresponse.core import Hamiltonian, Program, Spin
from pymolresponse.interfaces.pyscf import molecules
from pymolresponse.interfaces.pyscf.utils import occupations_from_pyscf_mol
from pymolresponse.properties import optrot

BC2H4_cation_HF_STO3G_RPA_singlet_nwchem = {
    # from minimal
    0.0: {
        "polar": np.array(
            [
                [16.9282686, -4.8000244, -0.7875224],
                [-4.8000245, 15.3007190, -2.7824200],
                [-0.7875224, -2.7824200, 19.3955636],
            ]
        )
    },
    # from orbeta
    0.001: {
        "polar": np.array(
            [
                [16.9283289, -4.8000606, -0.7875053],
                [-4.8000606, 15.3008509, -2.7825269],
                [-0.7875053, -2.7825269, 19.3957197],
            ]
        ),
        "orbeta": np.array(
            [[0.0684, -9.8683, -1.1279], [8.8646, -7.1416, 6.5276], [0.9846, 5.6305, -12.6550]]
        ),
    },
    # from orbeta
    0.0773178: {
        "polar": np.array(
            [
                [17.3117129, -5.0815177, -0.6234223],
                [-5.0815177, 16.4754295, -3.8455056],
                [-0.6234223, -3.8455056, 20.8275781],
            ]
        ),
        "orbeta": np.array(
            [
                [-0.0736, -9.0074, -1.3061],
                [10.7935, -13.5900, 14.8878],
                [-1.1477, 11.8707, -25.1217],
            ]
        ),
    },
    # from orbeta
    0.128347: {
        "polar": np.array(
            [
                [18.0091879, -4.7985368, -1.4998108],
                [-4.7985368, 11.1823569, 3.6436223],
                [-1.4998108, 3.6436223, 11.9881428],
            ]
        ),
        "orbeta": np.array(
            [
                [1.9247, -11.4465, 19.3223],
                [-3.6550, 18.6493, -101.8119],
                [16.9194, -35.1753, 99.3169],
            ]
        ),
    },
}


BC2H4_HF_STO3G_RPA_singlet_nwchem = {
    # from minimal
    0.0: {
        "polar": np.array(
            [
                [16.0702163, -4.9832488, -2.2293822],
                [-4.9832488, 13.1502378, 0.2844524],
                [-2.2293822, 0.2844524, 18.1993490],
            ]
        )
    },
    # from orbeta
    0.001: {
        "polar": np.array(
            [
                [16.0702789, -4.9832818, -2.2293882],
                [-4.9832818, 13.1502964, 0.2844393],
                [-2.2293882, 0.2844393, 18.1994070],
            ]
        ),
        "orbeta": np.array(
            [[-1.7067, -5.8002, -2.9465], [7.2404, 0.9278, 5.3503], [4.9573, -0.5056, -3.3741]]
        ),
    },
    # from orbeta
    0.0773178: {
        "polar": np.array(
            [
                [16.4963293, -5.2448779, -2.2476469],
                [-5.2448779, 13.6064856, 0.1637870],
                [-2.2476469, 0.1637869, 18.5823648],
            ]
        ),
        "orbeta": np.array(
            [[-1.4712, -5.3345, -4.2604], [6.4702, 0.3833, 8.8611], [5.6097, -0.4006, -5.0196]]
        ),
    },
    # from orbeta
    0.128347: {
        "polar": np.array(
            [
                [15.3394615, -2.0395326, -3.7937930],
                [-2.0395327, 7.7254042, 2.6104237],
                [-3.7937933, 2.6104240, 18.3446840],
            ]
        ),
        # Flipped the sign on the [2, 1] element.
        "orbeta": np.array(
            [[-17.0050, -8.6288, 53.2469], [33.1190, 7.7765, -90.7071], [-4.0315, -3.5893, 34.0186]]
        ),
    },
}


trithiolane_HF_STO3G_RPA_singlet = {
    0.0: {
        "polar": np.asarray(
            [
                [3.47899000e01, -1.77591000e-05, 2.24416000e-04],
                [-1.77591000e-05, 4.43286000e01, -1.91191000e-01],
                [2.24416000e-04, -1.91191000e-01, 1.68476000e01],
            ]
        ),
        "dipmag": np.asarray(
            [
                [-0.854380e01, -0.279465e-02, 0.693947e-03],
                [-0.537619e-02, 0.125637e02, -0.447617e01],
                [0.281657e-02, -0.985703e01, -0.506194e01],
            ]
        ),
    },
    0.0773178: {
        "polar": np.asarray(
            [
                [3.53972000e01, -8.48249000e-06, 2.05118000e-04],
                [-8.48249000e-06, 4.52552000e01, -2.17003000e-01],
                [2.05118000e-04, -2.17003000e-01, 1.70848000e01],
            ]
        ),
        "dipmag": np.asarray(
            [
                [-0.873446e01, -0.291464e-02, 0.749139e-03],
                [-0.546200e-02, 0.129058e02, -0.453107e01],
                [0.290515e-02, -0.105516e02, -0.537653e01],
            ]
        ),
    },
    0.128347: {
        "polar": np.asarray(
            [
                [3.65252000e01, -2.36748000e-06, 1.76171000e-04],
                [-2.36748000e-06, 4.70314000e01, -2.86868000e-01],
                [1.76171000e-04, -2.86868000e-01, 1.75452000e01],
            ]
        ),
        "dipmag": np.asarray(
            [
                [-0.908406e01, -0.315108e-02, 0.843565e-03],
                [-0.563037e-02, 0.135977e02, -0.466676e01],
                [0.306764e-02, -0.118891e02, -0.609620e01],
            ]
        ),
    },
}


def test_ORD_RPA_singlet_BC2H4_cation_HF_STO3G() -> None:
    ref = BC2H4_cation_HF_STO3G_RPA_singlet_nwchem

    pyscfmol = molecules.molecule_bc2h4_cation_sto3g()
    pyscfmol.build()

    mf = pyscf.scf.RHF(pyscfmol)
    mf.scf()

    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(pyscfmol, C)

    frequencies = [0.0, 0.001, 0.0773178, 0.128347]
    ord_solver = optrot.ORD(
        Program.PySCF,
        pyscfmol,
        cphf.CPHF(solvers.ExactInv(C, E, occupations)),
        C,
        E,
        occupations,
        frequencies=frequencies,
        do_dipvel=False,
    )
    ord_solver.form_operators()
    ord_solver.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    ord_solver.form_results()

    print("Polarizabilities")
    assert len(frequencies) == len(ord_solver.polarizabilities)
    thresh = 5.0e-4
    for idxf, frequency in enumerate(frequencies):
        ref_polar = ref[frequency]["polar"]
        res_polar = ord_solver.polarizabilities[idxf]
        abs_diff = abs(res_polar - ref_polar)
        print(idxf, frequency)
        print(res_polar)
        print(abs_diff)
        assert (abs_diff < thresh).all()

    print(r"\beta(\omega)")
    thresh = 5.0e-2
    for idxf, frequency in enumerate(frequencies):
        if "orbeta" in ref[frequency]:
            ref_beta = ref[frequency]["orbeta"]
            # TODO why no speed of light?
            # TODO why the (1/2)?
            res_beta = -(0.5 / frequency) * ord_solver.driver.results[idxf][3:6, 0:3]
            abs_diff = abs(res_beta - ref_beta)
            print(idxf, frequency)
            print(res_beta)
            print(abs_diff)
            assert (abs_diff < thresh).all()

    # from ecd import ECD
    # ecd = ECD(pyscfmol, C, E, occupations, do_dipvel=True, do_tda=False)
    # ecd.run()
    # ecd.form_results()
    # ord_solver.form_operators()
    # ord_solver.run()
    # ord_solver.form_results()
    # from constants import esuecd
    # prefac = -(2 / 3) / esuecd
    # for idxf, frequency in enumerate(frequencies):
    #     print(sum(prefac * ecd.rotational_strengths_diplen / ((frequency ** 2) * (ecd.solver.eigvals.real ** 2))))


def test_ORD_RPA_singlet_BC2H4_HF_STO3G() -> None:
    ref = BC2H4_HF_STO3G_RPA_singlet_nwchem

    pyscfmol = molecules.molecule_bc2h4_neutral_radical_sto3g()
    pyscfmol.build()

    mf = pyscf.scf.UHF(pyscfmol)
    mf.scf()

    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(pyscfmol, C)

    frequencies = [0.0, 0.001, 0.0773178, 0.128347]
    ord_solver = optrot.ORD(
        Program.PySCF,
        pyscfmol,
        cphf.CPHF(solvers.ExactInv(C, E, occupations)),
        C,
        E,
        occupations,
        frequencies=frequencies,
        do_dipvel=False,
    )
    ord_solver.form_operators()
    ord_solver.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    ord_solver.form_results()

    print("Polarizabilities")
    assert len(frequencies) == len(ord_solver.polarizabilities)
    thresh = 6.5e-4
    for idxf, frequency in enumerate(frequencies):
        ref_polar = ref[frequency]["polar"]
        res_polar = ord_solver.polarizabilities[idxf]
        abs_diff = abs(res_polar - ref_polar)
        print(idxf, frequency)
        print(res_polar)
        print(abs_diff)
        if frequency == 0.128347:
            thresh = 6.0e-3
        assert (abs_diff < thresh).all()

    print(r"\beta(\omega)")
    thresh = 0.09
    for idxf, frequency in enumerate(frequencies):
        if "orbeta" in ref[frequency]:
            ref_beta = ref[frequency]["orbeta"]
            # TODO why no speed of light?
            # TODO why the (1/2)?
            res_beta = -(0.5 / frequency) * ord_solver.driver.results[idxf][3:6, 0:3]
            abs_diff = abs(res_beta - ref_beta)
            print(idxf, frequency)
            print(res_beta)
            print(ref_beta)
            print(abs_diff)
            assert (abs_diff < thresh).all()


def test_ORD_RPA_singlet_trithiolane_HF_STO3G() -> None:
    ref = trithiolane_HF_STO3G_RPA_singlet

    pyscfmol = molecules.molecule_trithiolane_sto3g()
    pyscfmol.build()

    mf = pyscf.scf.RHF(pyscfmol)
    mf.scf()

    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(pyscfmol, C)

    frequencies = [0.0, 0.0773178, 0.128347]
    ord_solver = optrot.ORD(
        Program.PySCF,
        pyscfmol,
        cphf.CPHF(solvers.ExactInv(C, E, occupations)),
        C,
        E,
        occupations,
        frequencies=frequencies,
        do_dipvel=False,
    )
    ord_solver.form_operators()
    ord_solver.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    ord_solver.form_results()

    print("Polarizabilities")
    assert len(frequencies) == len(ord_solver.polarizabilities)
    thresh = 5.0e-4
    for idxf, frequency in enumerate(frequencies):
        ref_polar = ref[frequency]["polar"]
        res_polar = ord_solver.polarizabilities[idxf]
        abs_diff = abs(res_polar - ref_polar)
        print(idxf, frequency)
        print(res_polar)
        print(abs_diff)
        assert (abs_diff < thresh).all()

    # print('Electric dipole-magnetic dipole polarizabilities')
    # assert len(frequencies) == len(ord_solver.polarizabilities_lenmag)
    # # thresh
    # for idxf, frequency in enumerate(frequencies):
    #     if 'dipmag' in ref[frequency]:
    #         ref_dipmag = ref[frequency]['dipmag']
    #         res_dipmag = ord_solver.polarizabilities_lenmag[idxf]
    #         print(idxf, frequency)
    #         # print(ref_dipmag)
    #         # print(res_dipmag)
    #         print(res_dipmag / ref_dipmag)

    # print(r'\beta(\omega)')
    # thresh = 5.0e-2
    # for idxf, frequency in enumerate(frequencies):
    #     if 'orbeta' in ref[frequency]:
    #         ref_beta = ref[frequency]['orbeta']
    #         # why no speed of light? atomic units...
    #         # TODO why the (1/2)?
    #         res_beta = -(0.5 / frequency) * ord_solver.solver.results[idxf][3:6, 0:3]
    #         abs_diff = abs(res_beta - ref_beta)
    #         print(idxf, frequency)
    #         print(res_beta)
    #         print(abs_diff)
    #         assert (abs_diff < thresh).all()



import pyscf

from pymolresponse import solvers, td, utils
from pymolresponse.core import AO2MOTransformationType, Hamiltonian, Program, Spin
from pymolresponse.interfaces.pyscf.ao2mo import AO2MOpyscf
from pymolresponse.interfaces.pyscf.utils import occupations_from_pyscf_mol


def test_HF_both_singlet_HF_STO3G() -> None:
    mol = pyscf.gto.Mole()
    mol.verbose = 0
    mol.output = None

    # pylint: disable=bad-whitespace
    mol.atom = [["H", (0.0, 0.0, 0.917)], ["F", (0.0, 0.0, 0.0)]]
    mol.basis = "sto-3g"
    mol.build()

    mf = pyscf.scf.RHF(mol)
    mf.scf()

    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)
    solver_tda = solvers.ExactDiagonalizationSolverTDA(C, E, occupations)
    solver_tdhf = solvers.ExactDiagonalizationSolver(C, E, occupations)
    ao2mo = AO2MOpyscf(C, mol.verbose, mol)
    ao2mo.perform_rhf_partial()
    tei_mo = ao2mo.tei_mo
    solver_tda.tei_mo = tei_mo
    solver_tda.tei_mo_type = AO2MOTransformationType.partial
    solver_tdhf.tei_mo = tei_mo
    solver_tdhf.tei_mo_type = AO2MOTransformationType.partial
    driver_tda = td.TDA(solver_tda)
    driver_tdhf = td.TDHF(solver_tdhf)

    nroots = 5

    print("TDA using TDA()")
    driver_tda.run(
        hamiltonian=Hamiltonian.TDA, spin=Spin.singlet, program=Program.PySCF, program_obj=mol
    )
    excitation_energies_tda_using_tda = driver_tda.solver.eigvals[:nroots].real
    print("TDA using TDHF()")
    driver_tdhf.run(
        hamiltonian=Hamiltonian.TDA, spin=Spin.singlet, program=Program.PySCF, program_obj=mol
    )
    excitation_energies_tda_using_tdhf = driver_tdhf.solver.eigvals[:nroots].real
    print("RPA using TDHF()")
    driver_tdhf.run(
        hamiltonian=Hamiltonian.RPA, spin=Spin.singlet, program=Program.PySCF, program_obj=mol
    )
    excitation_energies_rpa = driver_tdhf.solver.eigvals[:nroots].real

    assert excitation_energies_tda_using_tda.shape == excitation_energies_tda_using_tdhf.shape
    assert excitation_energies_tda_using_tdhf.shape == excitation_energies_rpa.shape

    # There should be no difference in the TDA results regardless of
    # which implementation used.
    assert (excitation_energies_tda_using_tda - excitation_energies_tda_using_tdhf).all() == 0

    # Now compare against reference_data
    ref_tda = HF_neutral_singlet_HF_STO3G_CIS_qchem
    ref_rpa = HF_neutral_singlet_HF_STO3G_RPA_qchem

    thresh = 1.0e-7
    for i in range(nroots):
        abs_diff = abs(ref_tda["etenergies"][i] - excitation_energies_tda_using_tda[i])
        assert abs_diff < thresh

    thresh = 1.0e-7
    for i in range(nroots):
        abs_diff = abs(ref_rpa["etenergies"][i] - excitation_energies_rpa[i])
        assert abs_diff < thresh


HF_neutral_singlet_HF_STO3G_qchem = -98.5707799863

HF_neutral_singlet_HF_STO3G_CIS_qchem = {
    "etenergies": [
        -98.067181246814 - HF_neutral_singlet_HF_STO3G_qchem,
        -98.067181246768 - HF_neutral_singlet_HF_STO3G_qchem,
        -97.686596454655 - HF_neutral_singlet_HF_STO3G_qchem,
        -96.958042326818 - HF_neutral_singlet_HF_STO3G_qchem,
        -72.879307887356 - HF_neutral_singlet_HF_STO3G_qchem,
    ],
    "etoscslen": [0.0002003489, 0.0002003489, 0.9621809426, 0.0531137481, 0.0691994928],
}


HF_neutral_singlet_HF_STO3G_RPA_qchem = {
    "etenergies": [
        -98.068185050585 - HF_neutral_singlet_HF_STO3G_qchem,
        -98.068185050538 - HF_neutral_singlet_HF_STO3G_qchem,
        -97.703584999956 - HF_neutral_singlet_HF_STO3G_qchem,
        -96.962988495302 - HF_neutral_singlet_HF_STO3G_qchem,
        -72.879331844690 - HF_neutral_singlet_HF_STO3G_qchem,
    ],
    "etoscslen": [0.0001877054, 0.0001877054, 0.7777380206, 0.0322221420, 0.0686085799],
}

# TODO
#
# def test_LiH_cation_TDA_singlet_HF_STO3G() -> None:
#     from pymolresponse.interfaces.pyscf.molecules import molecule_lih_cation_sto3g

#     mol = molecule_lih_cation_sto3g()
#     mf = pyscf.scf.UHF(mol)
#     mf.scf()
#     C = utils.fix_mocoeffs_shape(mf.mo_coeff)
#     E = utils.fix_moenergies_shape(mf.mo_energy)
#     occupations = occupations_from_pyscf_mol(mol, C)
#     solver_tda = solvers.ExactDiagonalizationSolverTDA(C, E, occupations)
#     # TODO i thought this was part of the solver interface
#     ao2mo = AO2MOpyscf(C, mol.verbose, mol)
#     ao2mo.perform_uhf_partial()
#     tei_mo = ao2mo.tei_mo
#     solver_tda.tei_mo = tei_mo
#     driver_tda = td.TDA(solver_tda)
#     print("TDA using TDA()")
#     driver_tda.run(
#         hamiltonian=Hamiltonian.TDA, spin=Spin.singlet, program=Program.PySCF, program_obj=mol
#     )
#     print(driver_tda.solver.eigvals.real)



"""Tools for performing AO-to-MO transformations of two-electron integrals."""

from typing import Optional, Sequence

import numpy as np

from pymolresponse.utils import fix_mocoeffs_shape


class AO2MO:
    """Interface for performing AO-to-MO tranformations of two-electron integrals."""

    # TODO see
    # https://github.com/psi4/psi4numpy/blob/master/Tutorials/01_Psi4NumPy-Basics/1f_tensor-manipulation.ipynb

    def __init__(
        self,
        C: np.ndarray,
        occupations: Sequence[int],
        verbose: int = 1,
        I: Optional[np.ndarray] = None,  # noqa: E741
    ) -> None:
        self.C = fix_mocoeffs_shape(C)
        self.occupations = occupations
        self.verbose = verbose
        self.I = I

        self.nocc_alph, self.nvirt_alph, self.nocc_beta, self.nvirt_beta = occupations

        self.tei_mo = tuple()

    @staticmethod
    def transform(
        I: np.ndarray,  # noqa: E741
        C1: np.ndarray,
        C2: np.ndarray,
        C3: np.ndarray,
        C4: np.ndarray,  # noqa: E741
    ) -> np.ndarray:
        """
        Transforms the 4-index ERI I with the 4 transformation matrices C1 to C4.
        """
        nao = I.shape[0]
        MO = np.dot(C1.T, I.reshape(nao, -1)).reshape(C1.shape[1], nao, nao, nao)

        MO = np.einsum("qB,Aqrs->ABrs", C2, MO)
        MO = np.einsum("rC,ABrs->ABCs", C3, MO)
        MO = np.einsum("sD,ABCs->ABCD", C4, MO)
        return MO

    def perform_rhf_full(self) -> None:
        r"""Perform the transformation :math:`(\mu\nu|\lambda\sigma) \rightarrow (pq|rs)`."""
        tei_mo = self.transform(self.I, self.C[0], self.C[0], self.C[0], self.C[0])
        self.tei_mo = (tei_mo,)

    def perform_rhf_partial(self) -> None:
        r"""Perform the transformation :math:`(\mu\nu|\lambda\sigma) \rightarrow (ia|jb), (ij|ab)`."""
        norb = self.nocc_alph + self.nvirt_alph
        oa = slice(0, self.nocc_alph)
        va = slice(self.nocc_alph, norb)
        tei_mo_ovov = self.transform(
            self.I, self.C[0, :, oa], self.C[0, :, va], self.C[0, :, oa], self.C[0, :, va]
        )
        tei_mo_oovv = self.transform(
            self.I, self.C[0, :, oa], self.C[0, :, oa], self.C[0, :, va], self.C[0, :, va]
        )
        self.tei_mo = (tei_mo_ovov, tei_mo_oovv)

    def perform_uhf_full(self) -> None:
        r"""Perform the transformation :math:`(\mu\nu|\lambda\sigma) \rightarrow (p^{\alpha}q^{\alpha}|r^{\alpha}s^{\alpha}), (p^{\alpha}q^{\alpha}|r^{\beta}s^{\beta}), (p^{\beta}q^{\beta}|r^{\alpha}s^{\alpha}), (p^{\beta}q^{\beta}|r^{\beta}s^{\beta})`."""

    def perform_uhf_partial(self) -> None:
        r"""Perform the transformation :math:`(\mu\nu|\lambda\sigma) \rightarrow (i^{\alpha}a^{\alpha}|j^{\alpha}b^{\alpha}), (i^{\alpha}a^{\alpha}|j^{\beta}b^{\beta}), (i^{\beta}a^{\beta}|j^{\alpha}b^{\alpha}), (i^{\beta}a^{\beta}|j^{\beta}b^{\beta}), (i^{\alpha}j^{\alpha}|a^{\alpha}b^{\alpha}), (i^{\beta}j^{\beta}|a^{\beta}b^{\beta})`."""



from abc import ABC, abstractmethod
from enum import Enum, auto, unique
from typing import Any, Mapping, Optional, Tuple

from attr import attrib, attrs
from attr.validators import instance_of, optional

import numpy as np

STARS = "********"


@unique
class IntegralSymmetry(Enum):
    SYMMETRIC = auto()
    ANTISYMMETRIC = auto()


@attrs(frozen=True, slots=True)
class IntegralLabel:
    label: str = attrib(validator=instance_of(str))
    comp: Optional[int] = attrib(validator=optional(instance_of(int)), default=None)
    symmetry: IntegralSymmetry = attrib(
        validator=instance_of(IntegralSymmetry), default=IntegralSymmetry.SYMMETRIC
    )


class Integrals(ABC):
    def __init__(self) -> None:
        pass

    def integrals(self, label: IntegralLabel) -> np.ndarray:
        if label is None:
            raise RuntimeError
        return self._compute(label)

    @abstractmethod
    def _compute(self, label: IntegralLabel) -> np.ndarray:
        """Compute the integrals of some operator using some implementation."""


class JK(ABC):
    """Interface for computing Coulomb and exchange integrals pre-contracted with
    the density.
    """

    def __init__(self) -> None:
        pass

    @abstractmethod
    def compute_from_density(self, D: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute J and K from some density."""

    @abstractmethod
    def compute_from_mocoeffs(
        self, C_left: np.ndarray, C_right: Optional[np.ndarray] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Compute J and K from MO coefficients."""


def read_binary(binaryfilename: str) -> bytes:
    """Return the bytes present in the given binary file name."""

    with open(binaryfilename, "rb") as binaryfile:
        readbytes = binaryfile.read()

    return readbytes


def parse_aoproper(integralfilename: str) -> Mapping[str, Any]:
    """Parse the AOPROPER file generated by DALTON and save the integral
    matrices to disk if asked.
    """

    integral_dict = dict()

    # Specify the encoding explicitly, so there's no confusion.
    encoding = "utf-8"

    # There are a few labels we want to ignore.
    labels_to_ignore = ("HUCKOVLP", "HUCKEL", "HJPOPOVL", "EOFLABEL")

    # For now, we naively read the whole file into memory, split
    # directly on the delimiter, *then* iterate. This will eventually
    # have to be changed over to a more memory-efficient version in
    # order to deal with hundreds of basis functions (if we ever want
    # that).
    integralfile_bytes = read_binary(integralfilename)
    # Remove b' \x00\x00\x00********' -> 12 bytes
    integralfile_bytes = integralfile_bytes[12:]
    integralfile_records = integralfile_bytes.split(STARS.encode(encoding=encoding))

    for record in integralfile_records:
        full_label = record[8:24].decode(encoding=encoding)
        # The first part shows the structure of the matrix.
        #  SQUARE   -> stored as the full N^2 matrix
        #  SYMMETRI -> stored as lower triangle with N*(N+1)/2 elements,
        #              ordered by rows
        #  ANTISYMM -> stored as lower triangle with N*(N+1)/2 elements,
        #              ordered by rows, (make the top triangle negative?)
        # The second part is the actual label
        # that would appear in the output file.
        shape, label = full_label[:8].strip(), full_label[8:].strip()

        # There are a few labels we want to ignore.
        if not any(label in x for x in labels_to_ignore):
            # top: b'1 A  15 ANTISYMMXDIPVEL  \x00\x00\x00\xa8\x00\x00\x00'
            # end: b'\xa8\x00\x00\x00 \x00\x00\x00' -> (168, 32)
            integrals_as_bytes = record[32:-8]
            # pylint: disable=no-member
            integrals_tril = np.fromstring(integrals_as_bytes, dtype=np.double)
            # positive solution to x = n*(n+1)/2
            nbasis = int(0.5 * (-1 + np.sqrt(1 + (8 * len(integrals_tril)))))
            # print(shape, label, len(integrals_as_bytes), nbasis, integrals_tril.shape)

            # form the full "square" matrix representation
            integrals_square = np.zeros(shape=(nbasis, nbasis))
            tril_indices = np.tril_indices(nbasis)
            integrals_square[tril_indices] = integrals_tril
            diag = np.diag(integrals_square) * np.eye(nbasis)
            if shape == "SYMMETRI":
                integrals_square = -diag + integrals_square + integrals_square.T
            elif shape == "ANTISYMM":
                integrals_square = -diag + integrals_square - integrals_square.T
                # If the integrals are antisymmetrized, the whole
                # thing should sum to zero.
                asum = abs(np.sum(integrals_square))
                assert asum < 1.0e-10
            else:
                print("Shouldn't be here.")

            # Was the (anti)symmetrization done correctly?
            assert integrals_square[tril_indices].all() == integrals_tril.all()

            record_dict = {
                "label": label,
                "nbasis": nbasis,
                "shape": shape,
                "integrals": integrals_square,
            }
            integral_dict[label] = record_dict

    return integral_dict


def _form_rhs_geometric(C: np.ndarray, occupations, natoms, MO, mints):
    import psi4

    pC = psi4.core.Matrix.from_array(C)
    nocc, nvirt, _, _ = occupations
    norb = nocc + nvirt
    o = slice(0, nocc)
    v = slice(nocc, norb)
    cart = ["_X", "_Y", "_Z"]
    oei_dict = {"S": "OVERLAP", "T": "KINETIC", "V": "POTENTIAL"}

    # Fock matrix (MO)
    T = (C.T).dot(np.asarray(mints.ao_kinetic())).dot(C)
    V = (C.T).dot(np.asarray(mints.ao_potential())).dot(C)
    H = T + V
    J = np.einsum("pqii->pq", MO[:, :, o, o])
    K = np.einsum("piqi->pq", MO[:, o, :, o])
    F = H + (2 * J) - K

    deriv1 = dict()
    for atom in range(natoms):
        for key in oei_dict:
            deriv1_mat = mints.mo_oei_deriv1(oei_dict[key], atom, pC, pC)
            for p in range(3):
                map_key = key + str(atom) + cart[p]
                deriv1[map_key] = np.asarray(deriv1_mat[p])
    for atom in range(natoms):
        string = "TEI" + str(atom)
        deriv1_mat = mints.mo_tei_deriv1(atom, pC, pC, pC, pC)
        for p in range(3):
            map_key = string + cart[p]
            deriv1[map_key] = np.asarray(deriv1_mat[p])
    # deriv2 = dict()
    # for atom1 in range(natoms):
    #     for atom2 in range(atom1 + 1):
    #         for key in oei_dict:
    #             string = key + str(atom1) + str(atom2)
    #             deriv2_mat = mints.mo_oei_deriv2(oei_dict[key], atom1, atom2, pC, pC)
    #             pq = 0
    #             for p in range(3):
    #                 for q in range(3):
    #                     map_key = string + cart[p] + cart[q]
    #                     deriv2[map_key] = np.asarray(deriv2_mat[pq])
    #                     pq += 1
    # for atom1 in range(natoms):
    #     for atom2 in range(atom1 + 1):
    #         string = "TEI" + str(atom1) + str(atom2)
    #         deriv2_mat = mints.mo_tei_deriv2(atom1, atom2, pC, pC, pC, pC)
    #         pq = 0
    #         for p in range(3):
    #             for q in range(3):
    #                 map_key = string + cart[p] + cart[q]
    #                 deriv2[map_key] = np.asarray(deriv2_mat[pq])
    #                 pq += 1
    F_grad = dict()
    B = dict()
    for atom in range(natoms):
        for p in range(3):
            key = str(atom) + cart[p]
            contr1 = np.einsum("pqmm->pq", deriv1["TEI" + key][:, :, o, o])
            contr2 = np.einsum("pmmq->pq", deriv1["TEI" + key][:, o, o, :])
            F_grad[key] = deriv1["T" + key] + deriv1["V" + key] + (2 * contr1) - contr2
    for atom in range(natoms):
        for p in range(3):
            key = str(atom) + cart[p]
            contr1 = np.einsum("ia,ii->ia", deriv1["S" + key][o, v], F[o, o])
            contr2 = np.einsum("iamn,mn->ia", MO[o, v, o, o], deriv1["S" + key][o, o])
            contr3 = np.einsum("inma,mn->ia", MO[o, o, o, v], deriv1["S" + key][o, o])
            B[key] = contr1 - F_grad[key][o, v] + (2 * contr2) - contr3
    return B


if __name__ == "__main__":
    dalton_integrals = parse_aoproper("r_lih_hf_sto-3g/dalton_response_rpa_singlet/AOPROPER")
    from pymolresponse.interfaces.dalton.utils import dalton_label_to_operator

    labels = dalton_integrals.keys()
    for label in labels:
        print(dalton_label_to_operator(label))



from typing import List, TypedDict

import numpy as np

import pyscf

from pymolresponse import solvers, td, utils
from pymolresponse.core import Hamiltonian, Program, Spin
from pymolresponse.interfaces.pyscf import molecules
from pymolresponse.interfaces.pyscf.utils import occupations_from_pyscf_mol
from pymolresponse.properties import ecd


class ECDReference(TypedDict):
    etenergies: List[float]
    etoscslen: List[float]
    etoscsmix: List[float]
    etoscsvel: List[float]
    etrotstrlen: List[float]
    etrotstrvel: List[float]
    nroots: int


BC2H4_cation_HF_STO3G_RPA_singlet_nwchem: ECDReference = {
    "etenergies": [
        0.116938283,
        0.153688860,
        0.302306677,
        0.327380785,
        0.340637548,
        0.391151295,
        0.427233992,
        0.521916988,
        0.534473141,
        0.567055549,
    ],
    "etoscslen": [
        0.01116,
        0.00473,
        0.00458,
        0.09761,
        0.09190,
        0.03514,
        0.03804,
        0.28900,
        0.19604,
        0.43408,
    ],
    "etoscsmix": [
        0.0074981,
        0.0054776,
        0.0052432,
        0.0748931,
        0.0547656,
        0.0207675,
        0.0205230,
        0.1853662,
        0.1077085,
        0.2874928,
    ],
    "etoscsvel": [
        0.0069989,
        0.0076728,
        0.0092702,
        0.0575288,
        0.0327921,
        0.0132039,
        0.0113311,
        0.1195340,
        0.0602382,
        0.1923483,
    ],
    "etrotstrlen": [
        -77.6721763,
        -11.6203780,
        13.6253032,
        203.2296044,
        -2.7209904,
        14.1994514,
        -16.5542125,
        -101.6752655,
        76.2221837,
        -106.0751407,
    ],
    "etrotstrvel": [
        -50.5415342,
        -38.3324307,
        -13.0716770,
        153.6307152,
        -0.2283890,
        14.8708870,
        -9.8364102,
        -73.4642390,
        40.6989398,
        -70.9771590,
    ],
    "nroots": 10,
}


BC2H4_neutral_radical_HF_STO3G_RPA_singlet_nwchem: ECDReference = {
    "etenergies": [],
    "etoscslen": [],
    "etoscsmix": [],
    "etoscsvel": [],
    "etrotstrlen": [],
    "etrotstrvel": [],
    "nroots": 10,
}


BC2H4_cation_HF_STO3G_TDA_singlet_orca: ECDReference = {
    "etenergies": [
        0.125219,
        0.160132,
        0.304206,
        0.330029,
        0.346364,
        0.394249,
        0.428824,
        0.525677,
        0.540485,
        0.573778,
    ],
    "etoscslen": [
        0.017454719,
        0.007513348,
        0.004928672,
        0.111374395,
        0.117626606,
        0.048377204,
        0.044244183,
        0.180619359,
        0.349435586,
        0.492350077,
    ],
    "etoscsmix": [],
    "etoscsvel": [
        0.016456475,
        0.000843190,
        0.010314035,
        0.022709524,
        0.005049133,
        0.004114132,
        0.006144748,
        0.043759216,
        0.051361944,
        0.124348134,
    ],
    "etrotstrlen": [
        -88.56094,
        -12.35844,
        14.45593,
        209.05890,
        0.42486,
        23.08029,
        -17.22645,
        -96.09431,
        76.23472,
        -135.00060,
    ],
    "etrotstrvel": [],
    "nroots": 10,
}


def test_ECD_TDA_singlet_BC2H4_cation_HF_STO3G() -> None:
    ref = BC2H4_cation_HF_STO3G_TDA_singlet_orca
    nroots = ref["nroots"]

    mol = molecules.molecule_bc2h4_cation_sto3g()
    mol.build()

    mf = pyscf.scf.RHF(mol)
    mf.scf()

    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    ecd_dipvel_tda = ecd.ECD(
        Program.PySCF,
        mol,
        td.TDHF(solvers.ExactDiagonalizationSolver(C, E, occupations)),
        C,
        E,
        occupations,
        do_dipvel=True,
    )
    ecd_dipvel_tda.form_operators()
    ecd_dipvel_tda.run(hamiltonian=Hamiltonian.TDA, spin=Spin.singlet)
    ecd_dipvel_tda.form_results()

    print("excitation energies")
    ref_etenergies = np.array(ref["etenergies"])
    res_etenergies = ecd_dipvel_tda.driver.solver.eigvals.real[:nroots]
    print("ref, res")
    for refval, resval in zip(ref_etenergies, res_etenergies):
        print(refval, resval)
    # TODO this might be from ORCA, should use NWChem instead
    thresh = 1.0e-3
    for i in range(nroots):
        abs_diff = abs(ref_etenergies[i] - res_etenergies[i])
        assert abs_diff < thresh

    print("dipole (length) oscillator strengths")
    ref_etoscslen = np.array(ref["etoscslen"])
    res_etoscslen = ecd_dipvel_tda.driver.solver.operators[1].total_oscillator_strengths[:nroots]
    print("ref, res")
    for refval, resval in zip(ref_etoscslen, res_etoscslen):
        print(refval, resval)
    thresh = 1.0e-3
    # np.testing.assert_allclose(ref_etoscslen, res_etoscslen)
    for i in range(nroots):
        abs_diff = abs(ref_etoscslen[i] - res_etoscslen[i])
        assert abs_diff < thresh

    # TODO
    print("TODO dipole (mixed length/velocity) oscillator strengths")

    # TODO
    print("TODO dipole (velocity) oscillator strengths")

    print("rotatory strengths (length)")
    ref_etrotstrlen = np.array(ref["etrotstrlen"])
    res_etrotstrlen = ecd_dipvel_tda.rotational_strengths_diplen[:nroots]
    print("ref, res")
    for refval, resval in zip(ref_etrotstrlen, res_etrotstrlen):
        print(refval, resval)
    # TODO unlike other quantities, the error isn't uniformly
    # distributed among the roots; how should this be handled?
    thresh = 1.0e2
    for i in range(nroots):
        abs_diff = abs(ref_etrotstrlen[i] - res_etrotstrlen[i])
        assert abs_diff < thresh

    # print('rotatory strengths (velocity)')
    # ref_etrotstrvel = np.array(ref['etrotstrvel'])
    # res_etrotstrvel = ecd_dipvel_tda.rotational_strengths_dipvel[:nroots]
    # print('ref, res')
    # for refval, resval in zip(ref_etrotstrvel, res_etrotstrvel):
    #     print(refval, resval)
    # thresh = 1.0e-2
    # for i in range(nroots):
    #     abs_diff = abs(ref_etrotstrvel[i] - res_etrotstrvel[i])
    #     assert abs_diff < thresh

    # print(ecd_dipvel_tda.print_results_nwchem())
    # print(ecd_dipvel_tda.print_results_orca())
    # print(ecd_dipvel_tda.print_results_qchem())


def test_ECD_RPA_singlet_BC2H4_cation_HF_STO3G() -> None:
    ref = BC2H4_cation_HF_STO3G_RPA_singlet_nwchem
    nroots = ref["nroots"]

    mol = molecules.molecule_bc2h4_cation_sto3g()
    mol.build()

    mf = pyscf.scf.RHF(mol)
    mf.scf()

    C = utils.fix_mocoeffs_shape(mf.mo_coeff)
    E = utils.fix_moenergies_shape(mf.mo_energy)
    occupations = occupations_from_pyscf_mol(mol, C)

    ecd_dipvel_rpa = ecd.ECD(
        Program.PySCF,
        mol,
        td.TDHF(solvers.ExactDiagonalizationSolver(C, E, occupations)),
        C,
        E,
        occupations,
        do_dipvel=True,
    )
    ecd_dipvel_rpa.form_operators()
    ecd_dipvel_rpa.run(hamiltonian=Hamiltonian.RPA, spin=Spin.singlet)
    ecd_dipvel_rpa.form_results()

    print("excitation energies")
    ref_etenergies = np.array(ref["etenergies"])
    res_etenergies = ecd_dipvel_rpa.driver.solver.eigvals.real[:nroots]
    print("ref, res")
    for refval, resval in zip(ref_etenergies, res_etenergies):
        print(refval, resval)
    thresh = 2.5e-7
    for i in range(nroots):
        abs_diff = abs(ref_etenergies[i] - res_etenergies[i])
        assert abs_diff < thresh

    print("dipole (length) oscillator strengths")
    ref_etoscslen = np.array(ref["etoscslen"])
    res_etoscslen = ecd_dipvel_rpa.driver.solver.operators[1].total_oscillator_strengths[:nroots]
    print("ref, res")
    for refval, resval in zip(ref_etoscslen, res_etoscslen):
        print(refval, resval)
    thresh = 1.0e-5
    for i in range(nroots):
        abs_diff = abs(ref_etoscslen[i] - res_etoscslen[i])
        assert abs_diff < thresh

    # TODO
    print("TODO dipole (mixed length/velocity) oscillator strengths")

    # TODO
    print("TODO dipole (velocity) oscillator strengths")
    ref_etoscsvel = np.array(ref["etoscsvel"])  # noqa: F841
    res_etoscsvel = ecd_dipvel_rpa.driver.solver.operators[2].total_oscillator_strengths[:nroots]  # noqa: F841
    # print('ref, res')
    # for refval, resval in zip(ref_etoscsvel, res_etoscsvel):
    #     print(refval, resval)
    # print(ref_etoscsvel / res_etoscsvel)
    # print(res_etoscsvel / ref_etoscsvel)

    print("rotatory strengths (length)")
    ref_etrotstrlen = np.array(ref["etrotstrlen"])
    res_etrotstrlen = ecd_dipvel_rpa.rotational_strengths_diplen[:nroots]
    print("ref, res")
    for refval, resval in zip(ref_etrotstrlen, res_etrotstrlen):
        print(refval, resval)
    # TODO unlike other quantities, the error isn't uniformly
    # distributed among the roots; how should this be handled?
    thresh = 1.5e1
    for i in range(nroots):
        abs_diff = abs(ref_etrotstrlen[i] - res_etrotstrlen[i])
        assert abs_diff < thresh

    print("rotatory strengths (velocity)")
    ref_etrotstrvel = np.array(ref["etrotstrvel"])
    res_etrotstrvel = ecd_dipvel_rpa.rotational_strengths_dipvel[:nroots]
    print("ref, res")
    for refval, resval in zip(ref_etrotstrvel, res_etrotstrvel):
        print(refval, resval)
    thresh = 1.0e-2
    for i in range(nroots):
        abs_diff = abs(ref_etrotstrvel[i] - res_etrotstrvel[i])
        assert abs_diff < thresh

    # with open(os.path.join(refdir, 'BC2H4_cation', 'nwchem_singlet_rpa_velocity_root.str')) as fh:
    #     ref_str = fh.read()
    # res_str = ecd_dipvel_rpa.print_results_nwchem()
    # assert res_str == ref_str
    # print(ecd_dipvel_rpa.print_results_nwchem())
    # print(ecd_dipvel_rpa.print_results_orca())
    # print(ecd_dipvel_rpa.print_results_qchem())

    # tmom_diplen = ecd_dipvel_rpa.driver.solver.operators[1].transition_moments
    # tmom_dipvel = ecd_dipvel_rpa.driver.solver.operators[2].transition_moments
    # print(tmom_diplen)
    # print('dipole')
    # for i in range(nroots):
    #     print((2 / 3) * res_etenergies[i] * np.dot(tmom_diplen[i], tmom_diplen[i]))
    # print('mixed')
    # for i in range(nroots):
    #     print((2 / 3) * res_etenergies[i] * np.dot(tmom_diplen[i], tmom_dipvel[i]))

    # print('sum rule')
    # print('ref_etoscslen:', sum(ref_etoscslen))
    # print('res_etoscslen:', sum(res_etoscslen))
    # # print('ref_etoscsmix:', sum(ref_etoscsmix))
    # # print('res_etoscsmix:', sum(res_etoscsmix))
    # print('ref_etoscsvel:', sum(ref_etoscsvel))
    # print('res_etoscsvel:', sum(res_etoscsvel))


# TODO once UHF is done
# def test_ECD_RPA_singlet_BC2H4_neutral_radical_HF_STO3G() -> None:
#     mol = molecules.molecule_bc2h4_neutral_radical_sto3g()
#     mol.build()

#     mf = pyscf.scf.UHF(mol)
#     mf.scf()

#     C = utils.fix_mocoeffs_shape(mf.mo_coeff)



import numpy as np

import pyscf

from pymolresponse.ao2mo import AO2MO
from pymolresponse.interfaces.pyscf import molecules as molecules_pyscf
from pymolresponse.interfaces.pyscf.ao2mo import AO2MOpyscf
from pymolresponse.interfaces.pyscf.utils import occupations_from_pyscf_mol
from pymolresponse.utils import fix_mocoeffs_shape


def test_ao2mo_hand_against_pyscf_rhf_full() -> None:
    mol = molecules_pyscf.molecule_physicists_water_sto3g()
    mol.build()

    mf = pyscf.scf.RHF(mol)
    mf.kernel()
    C = fix_mocoeffs_shape(mf.mo_coeff)
    occupations = occupations_from_pyscf_mol(mol, C)
    nocc, nvirt, _, _ = occupations
    nmo = nocc + nvirt
    ntransforms = 1

    ao2mo = AO2MOpyscf(C, verbose=mol.verbose, pyscfmol=mol)
    ao2mo.perform_rhf_full()
    assert len(ao2mo.tei_mo) == ntransforms
    tei_mo_pyscf = ao2mo.tei_mo[0]  # ty: ignore[index-out-of-bounds]

    tei_ao = mol.intor("int2e_sph", aosym="s1")

    print("1. Use the class method explicitly.")

    tei_mo_hand = AO2MO.transform(tei_ao, C[0], C[0], C[0], C[0])

    assert tei_mo_pyscf.shape == tei_mo_hand.shape == (nmo, nmo, nmo, nmo)
    np.testing.assert_allclose(tei_mo_hand, tei_mo_pyscf, rtol=0, atol=1.0e-15)

    print("2. Use the class method normally.")

    ao2mo_method = AO2MO(C, occupations, verbose=mol.verbose, I=tei_ao)
    ao2mo_method.perform_rhf_full()
    assert len(ao2mo_method.tei_mo) == ntransforms
    tei_mo_method = ao2mo_method.tei_mo[0]  # ty: ignore[index-out-of-bounds]

    assert tei_mo_pyscf.shape == tei_mo_method.shape == (nmo, nmo, nmo, nmo)
    np.testing.assert_allclose(tei_mo_method, tei_mo_pyscf, rtol=0, atol=1.0e-15)


def test_ao2mo_hand_against_pyscf_rhf_partial() -> None:
    mol = molecules_pyscf.molecule_physicists_water_sto3g()
    mol.build()

    mf = pyscf.scf.RHF(mol)
    mf.kernel()
    C = fix_mocoeffs_shape(mf.mo_coeff)
    occupations = occupations_from_pyscf_mol(mol, C)
    nocc, nvirt, _, _ = occupations
    nmo = nocc + nvirt
    ntransforms = 2

    o = slice(0, nocc)
    v = slice(nocc, nmo)

    ao2mo = AO2MOpyscf(C, verbose=mol.verbose, pyscfmol=mol)
    ao2mo.perform_rhf_partial()
    assert len(ao2mo.tei_mo) == ntransforms
    tei_mo_ovov_pyscf = ao2mo.tei_mo[0]  # ty: ignore[index-out-of-bounds]
    tei_mo_oovv_pyscf = ao2mo.tei_mo[1]  # ty: ignore[index-out-of-bounds]

    tei_ao = mol.intor("int2e_sph", aosym="s1")

    print("1. Use the class method explicitly.")

    tei_mo_ovov_hand = AO2MO.transform(tei_ao, C[0, :, o], C[0, :, v], C[0, :, o], C[0, :, v])
    tei_mo_oovv_hand = AO2MO.transform(tei_ao, C[0, :, o], C[0, :, o], C[0, :, v], C[0, :, v])

    assert tei_mo_ovov_pyscf.shape == tei_mo_ovov_hand.shape == (nocc, nvirt, nocc, nvirt)
    assert tei_mo_oovv_pyscf.shape == tei_mo_oovv_hand.shape == (nocc, nocc, nvirt, nvirt)
    np.testing.assert_allclose(tei_mo_ovov_hand, tei_mo_ovov_pyscf, rtol=0, atol=1.0e-15)
    np.testing.assert_allclose(tei_mo_oovv_hand, tei_mo_oovv_pyscf, rtol=0, atol=1.0e-15)

    print("2. Use the class method normally.")

    ao2mo_method = AO2MO(C, occupations, verbose=mol.verbose, I=tei_ao)
    ao2mo_method.perform_rhf_partial()
    assert len(ao2mo_method.tei_mo) == ntransforms
    tei_mo_ovov_method = ao2mo_method.tei_mo[0]  # ty: ignore[index-out-of-bounds]
    tei_mo_oovv_method = ao2mo_method.tei_mo[1]  # ty: ignore[index-out-of-bounds]

    assert tei_mo_ovov_pyscf.shape == tei_mo_ovov_method.shape == (nocc, nvirt, nocc, nvirt)
    assert tei_mo_oovv_pyscf.shape == tei_mo_oovv_method.shape == (nocc, nocc, nvirt, nvirt)
    np.testing.assert_allclose(tei_mo_ovov_method, tei_mo_ovov_pyscf, rtol=0, atol=1.0e-15)
    np.testing.assert_allclose(tei_mo_oovv_method, tei_mo_oovv_pyscf, rtol=0, atol=1.0e-15)



"""Core types."""

from enum import Enum, unique


@unique
class AO2MOTransformationType(Enum):
    r"""For routines that perform AO-to-MO transformations, specify the kind of
    transformation.

    Starting from :math:`(\mu\nu|\lambda\sigma)`,

    - a 'partial' transformation will produce :math:`(ia|jb)` and :math:`(ij|ab)` if needed
    - a 'full' transformation will produce :math:`(pq|rs)`
    """

    partial = "partial"
    full = "full"


@unique
class Hamiltonian(Enum):
    """Specify which approximation for the orbital Hessian should be used.

    - RPA indicates the random phase approximation: the orbital Hessian is
      exact for the first-order polarization propagator.
    - TDA indicates the Tamm-Dancoff approximation: the B matrix is set to zero.

    Currently this is only capable of handling the usual first-order
    polarization propagator for CPHF and TDHF.
    """

    RPA = "rpa"
    TDA = "tda"


@unique
class Spin(Enum):
    """Specify whether the calculation should conserve spin (singlet) or flip spin
    (triplet).

    For response properties, this affects both the orbital Hessian and the
    perturbing operator. Note that these two specifications are currently
    independent!

    For excitation energies, this only affects the orbital Hessian and
    determines whether singlet or triplet excitation energies are calculated.
    """

    singlet = 0
    triplet = 1


@unique
class Program(Enum):
    """Specify which program should be used for a particular step."""

    PySCF = "pyscf"
    Psi4 = "psi4"
    DALTON = "dalton"
